# -*- generated by 1.0.12 -*-
import da
PatternExpr_212 = da.pat.TuplePattern([da.pat.ConstantPattern('Promise'), da.pat.BoundPattern('_BoundPattern215_'), da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern('a')])
PatternExpr_245 = da.pat.TuplePattern([da.pat.ConstantPattern('Promise'), da.pat.BoundPattern('_BoundPattern248_'), da.pat.FreePattern('n'), da.pat.FreePattern('v'), da.pat.FreePattern(None)])
PatternExpr_275 = da.pat.TuplePattern([da.pat.ConstantPattern('Promise'), da.pat.BoundPattern('_BoundPattern278_'), da.pat.BoundPattern('_BoundPattern279_'), da.pat.BoundPattern('_BoundPattern280_'), da.pat.FreePattern('a')])
PatternExpr_330 = da.pat.TuplePattern([da.pat.ConstantPattern('TwoAv'), da.pat.BoundPattern('_BoundPattern333_'), da.pat.BoundPattern('_BoundPattern334_'), da.pat.FreePattern('a')])
PatternExpr_417 = da.pat.TuplePattern([da.pat.ConstantPattern('TwoAv'), da.pat.FreePattern('n'), da.pat.FreePattern('v'), da.pat.FreePattern(None)])
PatternExpr_446 = da.pat.TuplePattern([da.pat.ConstantPattern('TwoAv'), da.pat.BoundPattern('_BoundPattern449_'), da.pat.BoundPattern('_BoundPattern450_'), da.pat.FreePattern('a')])
PatternExpr_474 = da.pat.TuplePattern([da.pat.ConstantPattern('TwoB'), da.pat.BoundPattern('_BoundPattern477_'), da.pat.BoundPattern('_BoundPattern478_')])
PatternExpr_509 = da.pat.TuplePattern([da.pat.ConstantPattern('Done')])
PatternExpr_514 = da.pat.BoundPattern('_BoundPattern515_')
PatternExpr_530 = da.pat.TuplePattern([da.pat.ConstantPattern('Prepare'), da.pat.FreePattern('n'), da.pat.FreePattern('p')])
PatternExpr_549 = da.pat.TuplePattern([da.pat.ConstantPattern('TwoAv'), da.pat.FreePattern('vpn'), da.pat.FreePattern('vv'), da.pat.SelfPattern()])
PatternExpr_601 = da.pat.TuplePattern([da.pat.ConstantPattern('OneC'), da.pat.FreePattern('n'), da.pat.FreePattern('v'), da.pat.FreePattern('p')])
PatternExpr_625 = da.pat.TuplePattern([da.pat.ConstantPattern('TwoAv'), da.pat.BoundPattern('_BoundPattern628_'), da.pat.FreePattern(None), da.pat.SelfPattern()])
PatternExpr_665 = da.pat.TuplePattern([da.pat.ConstantPattern('Promise'), da.pat.FreePattern('n'), da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)])
PatternExpr_703 = da.pat.TuplePattern([da.pat.ConstantPattern('Promise'), da.pat.BoundPattern('_BoundPattern706_'), da.pat.FreePattern('vn'), da.pat.FreePattern('vv'), da.pat.FreePattern(None)])
PatternExpr_733 = da.pat.TuplePattern([da.pat.ConstantPattern('Promise'), da.pat.BoundPattern('_BoundPattern736_'), da.pat.BoundPattern('_BoundPattern737_'), da.pat.BoundPattern('_BoundPattern738_'), da.pat.FreePattern('a')])
PatternExpr_481 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('TwoB'), da.pat.BoundPattern('_BoundPattern491_'), da.pat.BoundPattern('_BoundPattern492_')])])
PatternExpr_516 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern522_')]), da.pat.TuplePattern([da.pat.ConstantPattern('Done')])])
_config_object = {}
import sys

class Proposer(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._ProposerReceivedEvent_0 = []
        self._ProposerReceivedEvent_1 = []
        self._ProposerReceivedEvent_2 = []
        self._ProposerReceivedEvent_3 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_0', PatternExpr_212, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_1', PatternExpr_245, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_2', PatternExpr_275, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_3', PatternExpr_330, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, acceptors, quorumsize, f, nrounds, timeout, **rest_907):
        super().setup(acceptors=acceptors, quorumsize=quorumsize, f=f, nrounds=nrounds, timeout=timeout, **rest_907)
        self._state.acceptors = acceptors
        self._state.quorumsize = quorumsize
        self._state.f = f
        self._state.nrounds = nrounds
        self._state.timeout = timeout
        self._state.propNum = (0, self._id)
        self._state.propVal = self._id

    def run(self):
        count = 0
        while (count < self._state.nrounds):
            self.work()
            super()._label('prepare', block=False)
            self.send(('Prepare', self._state.propNum, self._id), to=self._state.acceptors)
            super()._label('_st_label_207', block=False)
            _st_label_207 = 0
            self._timer_start()
            while (_st_label_207 == 0):
                _st_label_207 += 1
                if (len({a for (_, _, (_ConstantPattern230_, _BoundPattern232_, _, _, a)) in self._ProposerReceivedEvent_0 if (_ConstantPattern230_ == 'Promise') if (_BoundPattern232_ == self._state.propNum)}) > self._state.quorumsize):
                    super()._label('propose', block=False)
                    (_, voted) = max(({(n, v) for (_, _, (_ConstantPattern264_, _BoundPattern266_, n, v, _)) in self._ProposerReceivedEvent_1 if (_ConstantPattern264_ == 'Promise') if (_BoundPattern266_ == self._state.propNum) if (len({a for (_, _, (_ConstantPattern293_, _BoundPattern295_, _BoundPattern296_, _BoundPattern297_, a)) in self._ProposerReceivedEvent_2 if (_ConstantPattern293_ == 'Promise') if (_BoundPattern295_ == self._state.propNum) if (_BoundPattern296_ == n) if (_BoundPattern297_ == v)}) > self._state.f)} | {(((- 1), self._id), self._state.propVal)}))
                    self.send(('OneC', self._state.propNum, voted, self._id), to=self._state.acceptors)
                    super()._label('_st_label_325', block=False)
                    _st_label_325 = 0
                    self._timer_start()
                    while (_st_label_325 == 0):
                        _st_label_325 += 1
                        if (len({a for (_, _, (_ConstantPattern347_, _BoundPattern349_, _BoundPattern350_, a)) in self._ProposerReceivedEvent_3 if (_ConstantPattern347_ == 'TwoAv') if (_BoundPattern349_ == self._state.propNum) if (_BoundPattern350_ == voted)}) > self._state.quorumsize):
                            super()._label('end', block=False)
                            self.output(('Succeeded proposing %s' % (voted,)))
                            count += 1
                            continue
                            _st_label_325 += 1
                        elif self._timer_expired:
                            self.output('Failed to Propose in time, retrying.')
                            _st_label_325 += 1
                        else:
                            super()._label('_st_label_325', block=True, timeout=self._state.timeout)
                            _st_label_325 -= 1
                    else:
                        if (_st_label_325 != 2):
                            continue
                    if (_st_label_325 != 2):
                        break
                    _st_label_207 += 1
                elif self._timer_expired:
                    self.output('Failed to Prepare in time, retrying.')
                    _st_label_207 += 1
                else:
                    super()._label('_st_label_207', block=True, timeout=self._state.timeout)
                    _st_label_207 -= 1
            else:
                if (_st_label_207 != 2):
                    continue
            if (_st_label_207 != 2):
                break
            self._state.propNum = ((self._state.propNum[0] + 1), self._id)
        self.send(('Done',), to=self._state.acceptors)

class Acceptor(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._AcceptorReceivedEvent_0 = []
        self._AcceptorReceivedEvent_1 = []
        self._AcceptorSentEvent_2 = []
        self._AcceptorReceivedEvent_3 = []
        self._AcceptorSentEvent_5 = []
        self._AcceptorSentEvent_7 = []
        self._AcceptorSentEvent_8 = []
        self._AcceptorReceivedEvent_9 = []
        self._AcceptorReceivedEvent_10 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_0', PatternExpr_417, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_1', PatternExpr_446, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_2', PatternExpr_474, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_3', PatternExpr_509, sources=[PatternExpr_514], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_4', PatternExpr_530, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Acceptor_handler_529]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_5', PatternExpr_549, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_6', PatternExpr_601, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Acceptor_handler_600]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_7', PatternExpr_625, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_8', PatternExpr_665, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_9', PatternExpr_703, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_10', PatternExpr_733, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, acceptors, proposers, quorumsize, f, **rest_907):
        super().setup(acceptors=acceptors, proposers=proposers, quorumsize=quorumsize, f=f, **rest_907)
        self._state.acceptors = acceptors
        self._state.proposers = proposers
        self._state.quorumsize = quorumsize
        self._state.f = f
        self._state.peers = (self._state.acceptors | self._state.proposers)

    def run(self):
        while True:
            super()._label('_st_label_414', block=False)
            a = v = n = None

            def ExistentialOpExpr_415():
                nonlocal a, v, n
                for (_, _, (_ConstantPattern435_, n, v, _)) in self._AcceptorReceivedEvent_0:
                    if (_ConstantPattern435_ == 'TwoAv'):
                        if ((len({a for (_, _, (_ConstantPattern463_, _BoundPattern465_, _BoundPattern466_, a)) in self._AcceptorReceivedEvent_1 if (_ConstantPattern463_ == 'TwoAv') if (_BoundPattern465_ == n) if (_BoundPattern466_ == v)}) > self._state.quorumsize) and (not PatternExpr_481.match_iter(self._AcceptorSentEvent_2, _BoundPattern491_=n, _BoundPattern492_=v, SELF_ID=self._id))):
                            return True
                return False
            p = None

            def UniversalOpExpr_501():
                nonlocal p
                for p in self._state.proposers:
                    if (not PatternExpr_516.match_iter(self._AcceptorReceivedEvent_3, _BoundPattern522_=p, SELF_ID=self._id)):
                        return False
                return True
            _st_label_414 = 0
            while (_st_label_414 == 0):
                _st_label_414 += 1
                if ExistentialOpExpr_415():
                    self.send(('TwoB', n, v), to=self._state.peers)
                    _st_label_414 += 1
                elif UniversalOpExpr_501():
                    break
                    _st_label_414 += 1
                else:
                    super()._label('_st_label_414', block=True)
                    _st_label_414 -= 1
            else:
                if (_st_label_414 != 2):
                    continue
            if (_st_label_414 != 2):
                break

    def maxpromised(self):
        return max(({n for (_, _, (_ConstantPattern683_, n, _, _, _)) in self._AcceptorSentEvent_8 if (_ConstantPattern683_ == 'Promise')} | {((- 2), self._id)}))

    def islegal(self, n, v):
        voted = {(vn, vv) for (_, _, (_ConstantPattern722_, _BoundPattern724_, vn, vv, _)) in self._AcceptorReceivedEvent_9 if (_ConstantPattern722_ == 'Promise') if (_BoundPattern724_ == n) if (len({a for (_, _, (_ConstantPattern751_, _BoundPattern753_, _BoundPattern754_, _BoundPattern755_, a)) in self._AcceptorReceivedEvent_10 if (_ConstantPattern751_ == 'Promise') if (_BoundPattern753_ == n) if (_BoundPattern754_ == vn) if (_BoundPattern755_ == vv)}) > self._state.f)}
        if (voted and (not (max(voted)[1] is None))):
            return (v == max(voted)[1])
        else:
            return True

    def _Acceptor_handler_529(self, n, p):
        if (n > self.maxpromised()):
            (vn, vv) = max(({(vpn, vv) for (_, _, (_ConstantPattern568_, vpn, vv, _ConstantPattern572_)) in self._AcceptorSentEvent_5 if (_ConstantPattern568_ == 'TwoAv') if (_ConstantPattern572_ == self._id)} | {(((- 1), self._id), None)}))
            self.send(('Promise', n, vn, vv, self._id), to=self._state.peers)
    _Acceptor_handler_529._labels = None
    _Acceptor_handler_529._notlabels = None

    def _Acceptor_handler_600(self, n, v, p):

        def ExistentialOpExpr_623():
            for (_, _, (_ConstantPattern642_, _BoundPattern644_, _, _ConstantPattern646_)) in self._AcceptorSentEvent_7:
                if (_ConstantPattern642_ == 'TwoAv'):
                    if (_BoundPattern644_ == n):
                        if (_ConstantPattern646_ == self._id):
                            if True:
                                return True
            return False
        if ((n >= self.maxpromised()) and self.islegal(n, v) and (not ExistentialOpExpr_623())):
            self.send(('TwoAv', n, v, self._id), to=self._state.peers)
    _Acceptor_handler_600._labels = None
    _Acceptor_handler_600._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        nproposers = (int(sys.argv[1]) if (len(sys.argv) > 1) else 5)
        nacceptors = (int(sys.argv[2]) if (len(sys.argv) > 2) else 10)
        nrounds = (int(sys.argv[3]) if (len(sys.argv) > 3) else 1)
        timeout = (int(sys.argv[4]) if (len(sys.argv) > 4) else 1)
        f = int(((nacceptors - 1) / 3))
        quorum = int(((nacceptors / 2) + f))
        acceptors = self.new(Acceptor, num=nacceptors)
        proposers = self.new(Proposer, num=nproposers)
        self._setup(acceptors, (acceptors, proposers, quorum, f))
        self._setup(proposers, (acceptors, quorum, f, nrounds, timeout))
        self._start(acceptors)
        self._start(proposers)



# -*- generated by 1.0.12 -*-
import da
PatternExpr_287 = da.pat.TuplePattern([da.pat.ConstantPattern('token'), da.pat.FreePattern('Q1'), da.pat.FreePattern('LN1')])
PatternExpr_304 = da.pat.TuplePattern([da.pat.ConstantPattern('request'), da.pat.FreePattern('p'), da.pat.FreePattern('n')])
PatternExpr_352 = da.pat.TuplePattern([da.pat.ConstantPattern('token'), da.pat.FreePattern(None), da.pat.FreePattern(None)])
PatternExpr_375 = da.pat.TuplePattern([da.pat.ConstantPattern('token'), da.pat.FreePattern(None), da.pat.FreePattern('LN1')])
PatternExpr_399 = da.pat.TuplePattern([da.pat.ConstantPattern('token'), da.pat.FreePattern(None), da.pat.FreePattern('LN2')])
PatternExpr_464 = da.pat.TuplePattern([da.pat.ConstantPattern('Done')])
PatternExpr_469 = da.pat.BoundPattern('_BoundPattern470_')
PatternExpr_471 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern477_')]), da.pat.TuplePattern([da.pat.ConstantPattern('Done')])])
_config_object = {}
import sys

class P(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._PSentEvent_2 = []
        self._PReceivedEvent_3 = []
        self._PSentEvent_4 = []
        self._PReceivedEvent_5 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_0', PatternExpr_287, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_286]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_1', PatternExpr_304, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_303]), da.pat.EventPattern(da.pat.SentEvent, '_PSentEvent_2', PatternExpr_352, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_3', PatternExpr_375, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_PSentEvent_4', PatternExpr_399, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_5', PatternExpr_464, sources=[PatternExpr_469], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, ps, orig_token, nrounds, **rest_559):
        super().setup(ps=ps, orig_token=orig_token, nrounds=nrounds, **rest_559)
        self._state.ps = ps
        self._state.orig_token = orig_token
        self._state.nrounds = nrounds
        self._state.RN = dict(((p, 0) for p in self._state.ps))
        self._state.Q = []
        self._state.LN = dict(((p, 0) for p in self._state.ps))

    def run(self):

        def anounce():
            self.output('In cs!')
        if self.haveToken():
            self.output("I'm lucky!")
        for i in range(self._state.nrounds):
            self.cs(anounce)
        self.send(('Done',), to=self._state.ps)
        super()._label('_st_label_455', block=False)
        p = None

        def UniversalOpExpr_456():
            nonlocal p
            for p in self._state.ps:
                if (not PatternExpr_471.match_iter(self._PReceivedEvent_5, _BoundPattern477_=p, SELF_ID=self._id)):
                    return False
            return True
        _st_label_455 = 0
        while (_st_label_455 == 0):
            _st_label_455 += 1
            if UniversalOpExpr_456():
                _st_label_455 += 1
            else:
                super()._label('_st_label_455', block=True)
                _st_label_455 -= 1
        self.output('Done!')

    def cs(self, task):
        super()._label('request', block=False)
        if (not self.haveToken()):
            self._state.RN[self._id] += 1
            self.send(('request', self._id, self._state.RN[self._id]), to=self._state.ps)
            super()._label('_st_label_229', block=False)
            _st_label_229 = 0
            while (_st_label_229 == 0):
                _st_label_229 += 1
                if self.haveToken():
                    _st_label_229 += 1
                else:
                    super()._label('_st_label_229', block=True)
                    _st_label_229 -= 1
        task()
        self._state.LN[self._id] = self._state.RN[self._id]
        self._state.Q.extend([p for p in self._state.ps if (not (p in self._state.Q)) if (self._state.RN[p] == (self._state.LN[p] + 1))])
        if (len(self._state.Q) > 0):
            p = self._state.Q.pop()
            self.send(('token', self._state.Q, self._state.LN), to=p)

    def haveToken(self):

        def ExistentialOpExpr_350():
            for (_, _, (_ConstantPattern367_, _, _)) in self._PSentEvent_2:
                if (_ConstantPattern367_ == 'token'):
                    if True:
                        return True
            return False
        LN1 = LN2 = None

        def ExistentialOpExpr_373():
            nonlocal LN1, LN2
            for (_, _, (_ConstantPattern391_, _, LN1)) in self._PReceivedEvent_3:
                if (_ConstantPattern391_ == 'token'):

                    def ExistentialOpExpr_397(LN1):
                        nonlocal LN2
                        for (_, _, (_ConstantPattern415_, _, LN2)) in self._PSentEvent_4:
                            if (_ConstantPattern415_ == 'token'):
                                if (LN2[self._id] > LN1[self._id]):
                                    return True
                        return False
                    if (not ExistentialOpExpr_397(LN1=LN1)):
                        return True
            return False
        return ((self._state.orig_token and (not ExistentialOpExpr_350())) or ExistentialOpExpr_373())

    def _P_handler_286(self, Q1, LN1):
        self._state.Q = Q1
        self._state.LN = LN1
    _P_handler_286._labels = None
    _P_handler_286._notlabels = None

    def _P_handler_303(self, p, n):
        self._state.RN[p] = max((self._state.RN[p], n))
        if (self.haveToken() and (self._state.RN[p] == (self._state.LN[p] + 1))):
            self.send(('token', self._state.Q, self._state.LN), to=p)
    _P_handler_303._labels = None
    _P_handler_303._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        nprocs = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        nrounds = (int(sys.argv[2]) if (len(sys.argv) > 2) else 1)
        ps = self.new(P, num=nprocs)
        lucky = ps.pop()
        self._setup(ps, ((ps | {lucky}), False, nrounds))
        self._setup({lucky}, ((ps | {lucky}), True, nrounds))
        self._start((ps | {lucky}))



# runtime package

from da import common, api, pattern as pat, compiler, sim
from da.sim import DistProcess, NodeProcess
from da.common import __version__

import_da = api.import_da
__all__ = ["__version__", "pat", "api", "compiler",
           "DistProcess", "NodeProcess",
           "import_da"]

for name in common.api_registry.keys():
    globals()[name] = common.api_registry[name]
    __all__.append(name)



# -*- generated by 1.0.12 -*-
import da
PatternExpr_175 = da.pat.TuplePattern([da.pat.ConstantPattern('Y')])
PatternExpr_185 = da.pat.TuplePattern([da.pat.ConstantPattern('Z')])
_config_object = {}

class P(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_0', PatternExpr_175, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_174]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_1', PatternExpr_185, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_184])])

    def setup(self, **rest_290):
        super().setup(**rest_290)
        self._state.d = 1
        self._state.x = 1

    def run(self):
        super()._label('l', block=False)
        _st_label_195 = 0
        while True:
            if (_st_label_195 == 2):
                break
            elif (_st_label_195 == 1):
                super()._label('l', block=True)
            _st_label_195 = 1
            if (self._state.d == 1):
                self.output('d is 1.')
            elif (self._state.x == 1):
                self.output('x is 1.')
            else:
                self.output('neither d nor x is 1.')
                _st_label_195 += 1

    def _P_handler_174(self):
        self._state.d = 0
    _P_handler_174._labels = None
    _P_handler_174._notlabels = None

    def _P_handler_184(self):
        self._state.x = 0
    _P_handler_184._labels = None
    _P_handler_184._notlabels = None

class Q(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def setup(self, ps, **rest_290):
        super().setup(ps=ps, **rest_290)
        self._state.ps = ps
        pass

    def run(self):
        self.output("Sending 'X'.")
        self.send(('X',), to=self._state.ps)
        super()._label('_st_label_228', block=False)
        _st_label_228 = 0
        self._timer_start()
        while (_st_label_228 == 0):
            _st_label_228 += 1
            if False:
                _st_label_228 += 1
            elif self._timer_expired:
                _st_label_228 += 1
            else:
                super()._label('_st_label_228', block=True, timeout=2)
                _st_label_228 -= 1
        self.output("Sending 'X'.")
        self.send(('X',), to=self._state.ps)
        super()._label('_st_label_239', block=False)
        _st_label_239 = 0
        self._timer_start()
        while (_st_label_239 == 0):
            _st_label_239 += 1
            if False:
                _st_label_239 += 1
            elif self._timer_expired:
                _st_label_239 += 1
            else:
                super()._label('_st_label_239', block=True, timeout=2)
                _st_label_239 -= 1
        self.output("Sending 'Y'.")
        self.send(('Y',), to=self._state.ps)
        super()._label('_st_label_250', block=False)
        _st_label_250 = 0
        self._timer_start()
        while (_st_label_250 == 0):
            _st_label_250 += 1
            if False:
                _st_label_250 += 1
            elif self._timer_expired:
                _st_label_250 += 1
            else:
                super()._label('_st_label_250', block=True, timeout=2)
                _st_label_250 -= 1
        self.output("Sending 'Z'.")
        self.send(('Z',), to=self._state.ps)
        while False:
            self.output('Fail!')

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        ps = self.new(P, [], 1)
        qs = self.new(Q, [ps], 1)
        self._start(ps)
        self._start(qs)



# -*- generated by 1.0.12 -*-
import da
PatternExpr_176 = da.pat.TuplePattern([da.pat.ConstantPattern('Ping')])
PatternExpr_181 = da.pat.FreePattern('p')
PatternExpr_227 = da.pat.TuplePattern([da.pat.ConstantPattern('Pong')])
PatternExpr_232 = da.pat.FreePattern('rclk')
PatternExpr_252 = da.pat.TuplePattern([da.pat.ConstantPattern('Pong')])
_config_object = {}
import sys

class Pong(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PongReceivedEvent_0', PatternExpr_176, sources=[PatternExpr_181], destinations=None, timestamps=None, record_history=None, handlers=[self._Pong_handler_175])])

    def setup(self, total_pings, **rest_324):
        super().setup(total_pings=total_pings, **rest_324)
        self._state.total_pings = total_pings
        pass

    def run(self):
        super()._label('_st_label_169', block=False)
        _st_label_169 = 0
        while (_st_label_169 == 0):
            _st_label_169 += 1
            if (self._state.total_pings == 0):
                _st_label_169 += 1
            else:
                super()._label('_st_label_169', block=True)
                _st_label_169 -= 1

    def _Pong_handler_175(self, p):
        self.output('Pinged')
        self.send(('Pong',), to=p)
        self._state.total_pings -= 1
    _Pong_handler_175._labels = None
    _Pong_handler_175._notlabels = None

class Ping(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._PingReceivedEvent_0 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PingReceivedEvent_0', PatternExpr_227, sources=None, destinations=None, timestamps=[PatternExpr_232], record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PingReceivedEvent_1', PatternExpr_252, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Ping_handler_251])])

    def setup(self, p, nrounds, **rest_324):
        super().setup(p=p, nrounds=nrounds, **rest_324)
        self._state.p = p
        self._state.nrounds = nrounds
        pass

    def run(self):
        for i in range(self._state.nrounds):
            clk = self.logical_clock()
            self.send(('Ping',), to=self._state.p)
            super()._label('_st_label_224', block=False)
            rclk = None

            def ExistentialOpExpr_225():
                nonlocal rclk
                for (_, (rclk, _, _), (_ConstantPattern243_,)) in self._PingReceivedEvent_0:
                    if (_ConstantPattern243_ == 'Pong'):
                        if (rclk > clk):
                            return True
                return False
            _st_label_224 = 0
            while (_st_label_224 == 0):
                _st_label_224 += 1
                if ExistentialOpExpr_225():
                    _st_label_224 += 1
                else:
                    super()._label('_st_label_224', block=True)
                    _st_label_224 -= 1
            else:
                if (_st_label_224 != 2):
                    continue
            if (_st_label_224 != 2):
                break

    def _Ping_handler_251(self):
        self.output('Ponged.')
    _Ping_handler_251._labels = None
    _Ping_handler_251._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])
    _config_object = {'clock': 'Lamport'}

    def run(self):
        nrounds = (int(sys.argv[1]) if (len(sys.argv) > 1) else 3)
        npings = (int(sys.argv[2]) if (len(sys.argv) > 2) else 3)
        pong = self.new(Pong, [(nrounds * npings)], num=1)
        ping = self.new(Ping, num=npings)
        self._setup(ping, (pong, nrounds))
        self._start(pong)
        self._start(ping)



# -*- generated by 1.0.12 -*-
import da
PatternExpr_216 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern219_'), da.pat.FreePattern(None)])
PatternExpr_223 = da.pat.FreePattern('a')
PatternExpr_251 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern254_'), da.pat.TuplePattern([da.pat.FreePattern('n2'), da.pat.FreePattern('v')])])
PatternExpr_282 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern285_'), da.pat.TuplePattern([da.pat.FreePattern('n2'), da.pat.FreePattern(None)])])
PatternExpr_318 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern321_'), da.pat.FreePattern(None)])
PatternExpr_325 = da.pat.FreePattern('a')
PatternExpr_366 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_371 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
PatternExpr_411 = da.pat.TuplePattern([da.pat.ConstantPattern('prepare'), da.pat.FreePattern('n')])
PatternExpr_418 = da.pat.FreePattern('p')
PatternExpr_424 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.FreePattern('n2'), da.pat.FreePattern(None)])
PatternExpr_453 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.FreePattern('n'), da.pat.FreePattern('v')])
PatternExpr_479 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.FreePattern('n'), da.pat.FreePattern(None)])
PatternExpr_514 = da.pat.TuplePattern([da.pat.ConstantPattern('accept'), da.pat.FreePattern('n'), da.pat.FreePattern('v')])
PatternExpr_527 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.FreePattern('n2'), da.pat.FreePattern(None)])
PatternExpr_563 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_568 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
PatternExpr_611 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.FreePattern('n'), da.pat.FreePattern('v')])
PatternExpr_637 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.BoundPattern('_BoundPattern640_'), da.pat.BoundPattern('_BoundPattern641_')])
PatternExpr_644 = da.pat.FreePattern('a')
PatternExpr_789 = da.pat.TuplePattern([da.pat.ConstantPattern('learned')])
PatternExpr_794 = da.pat.BoundPattern('_BoundPattern795_')
PatternExpr_796 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern802_')]), da.pat.TuplePattern([da.pat.ConstantPattern('learned')])])
_config_object = {}
import sys
from random import randint
TIMEOUT = 1

class Proposer(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._ProposerReceivedEvent_0 = []
        self._ProposerReceivedEvent_1 = []
        self._ProposerReceivedEvent_2 = []
        self._ProposerReceivedEvent_3 = []
        self._ProposerReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_0', PatternExpr_216, sources=[PatternExpr_223], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_1', PatternExpr_251, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_2', PatternExpr_282, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_3', PatternExpr_318, sources=[PatternExpr_325], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_4', PatternExpr_366, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, acceptors, **rest_815):
        super().setup(acceptors=acceptors, **rest_815)
        self._state.acceptors = acceptors
        self._state.n = None
        self._state.majority = self._state.acceptors

    def run(self):
        while (not PatternExpr_371.match_iter(self._ProposerReceivedEvent_4, SELF_ID=self._id)):
            self.to_consent()
        self.output('terminating')

    def to_consent(self):
        self._state.n = ((0, self._id) if (self._state.n == None) else ((self._state.n[0] + 1), self._id))
        self.send(('prepare', self._state.n), to=self._state.majority)
        super()._label('_st_label_211', block=False)
        _st_label_211 = 0
        self._timer_start()
        while (_st_label_211 == 0):
            _st_label_211 += 1
            if (len({a for (_, (_, _, a), (_ConstantPattern234_, _BoundPattern236_, _)) in self._ProposerReceivedEvent_0 if (_ConstantPattern234_ == 'respond') if (_BoundPattern236_ == self._state.n)}) > (len(self._state.acceptors) / 2)):
                v = self.anyof(({v for (_, _, (_ConstantPattern270_, _BoundPattern272_, (n2, v))) in self._ProposerReceivedEvent_1 if (_ConstantPattern270_ == 'respond') if (_BoundPattern272_ == self._state.n) if (n2 == max({n2 for (_, _, (_ConstantPattern299_, _BoundPattern301_, (n2, _))) in self._ProposerReceivedEvent_2 if (_ConstantPattern299_ == 'respond') if (_BoundPattern301_ == self._state.n)}))} or {randint(1, 100)}))
                responded = {a for (_, (_, _, a), (_ConstantPattern336_, _BoundPattern338_, _)) in self._ProposerReceivedEvent_3 if (_ConstantPattern336_ == 'respond') if (_BoundPattern338_ == self._state.n)}
                self.send(('accept', self._state.n, v), to=responded)
                self.debug('### chose', self._state.n, v)
                _st_label_211 += 1
            elif self._timer_expired:
                self.output('failed proposal number', self._state.n)
                _st_label_211 += 1
            else:
                super()._label('_st_label_211', block=True, timeout=TIMEOUT)
                _st_label_211 -= 1

    def anyof(self, s):
        return (next(iter(s)) if s else None)

class Acceptor(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._AcceptorSentEvent_1 = []
        self._AcceptorSentEvent_2 = []
        self._AcceptorSentEvent_3 = []
        self._AcceptorSentEvent_5 = []
        self._AcceptorReceivedEvent_6 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_0', PatternExpr_411, sources=[PatternExpr_418], destinations=None, timestamps=None, record_history=None, handlers=[self._Acceptor_handler_410]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_1', PatternExpr_424, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_2', PatternExpr_453, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_3', PatternExpr_479, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_4', PatternExpr_514, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Acceptor_handler_513]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_5', PatternExpr_527, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_6', PatternExpr_563, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, learners, **rest_815):
        super().setup(learners=learners, **rest_815)
        self._state.learners = learners
        pass

    def run(self):
        super()._label('_st_label_560', block=False)
        _st_label_560 = 0
        while (_st_label_560 == 0):
            _st_label_560 += 1
            if PatternExpr_568.match_iter(self._AcceptorReceivedEvent_6, SELF_ID=self._id):
                _st_label_560 += 1
            else:
                super()._label('_st_label_560', block=True)
                _st_label_560 -= 1
        self.output('terminating')

    def anyof(self, s):
        "return any element of set s if s is not empty or 'None' otherwise"
        return (next(iter(s)) if s else None)

    def _Acceptor_handler_410(self, n, p):
        n2 = None

        def UniversalOpExpr_422():
            nonlocal n2
            for (_, _, (_ConstantPattern440_, n2, _)) in self._AcceptorSentEvent_1:
                if (_ConstantPattern440_ == 'respond'):
                    if (not (n > n2)):
                        return False
            return True
        if UniversalOpExpr_422():
            maxprop = self.anyof({(n, v) for (_, _, (_ConstantPattern469_, n, v)) in self._AcceptorSentEvent_2 if (_ConstantPattern469_ == 'accepted') if (n == max({n for (_, _, (_ConstantPattern494_, n, _)) in self._AcceptorSentEvent_3 if (_ConstantPattern494_ == 'accepted')}))})
            self.send(('respond', n, maxprop), to=p)
    _Acceptor_handler_410._labels = None
    _Acceptor_handler_410._notlabels = None

    def _Acceptor_handler_513(self, n, v):
        n2 = None

        def ExistentialOpExpr_525():
            nonlocal n2
            for (_, _, (_ConstantPattern543_, n2, _)) in self._AcceptorSentEvent_5:
                if (_ConstantPattern543_ == 'respond'):
                    if (n2 > n):
                        return True
            return False
        if (not ExistentialOpExpr_525()):
            self.send(('accepted', n, v), to=self._state.learners)
    _Acceptor_handler_513._labels = None
    _Acceptor_handler_513._notlabels = None

class Learner(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._LearnerReceivedEvent_0 = []
        self._LearnerReceivedEvent_1 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_LearnerReceivedEvent_0', PatternExpr_611, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_LearnerReceivedEvent_1', PatternExpr_637, sources=[PatternExpr_644], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, acceptors, **rest_815):
        super().setup(acceptors=acceptors, **rest_815)
        self._state.acceptors = acceptors
        pass

    def run(self):
        self.learn()
        self.output('terminating')
        self.send(('learned',), to=self.nodeof(self._id))

    def learn(self):
        super()._label('_st_label_608', block=False)
        v = n = a = None

        def ExistentialOpExpr_609():
            nonlocal v, n, a
            for (_, _, (_ConstantPattern628_, n, v)) in self._LearnerReceivedEvent_0:
                if (_ConstantPattern628_ == 'accepted'):
                    if (len({a for (_, (_, _, a), (_ConstantPattern655_, _BoundPattern657_, _BoundPattern658_)) in self._LearnerReceivedEvent_1 if (_ConstantPattern655_ == 'accepted') if (_BoundPattern657_ == n) if (_BoundPattern658_ == v)}) > (len(self._state.acceptors) / 2)):
                        return True
            return False
        _st_label_608 = 0
        self._timer_start()
        while (_st_label_608 == 0):
            _st_label_608 += 1
            if ExistentialOpExpr_609():
                self.output('learned', n, v)
                _st_label_608 += 1
            elif self._timer_expired:
                self.output('failed learning anything')
                _st_label_608 += 1
            else:
                super()._label('_st_label_608', block=True, timeout=(TIMEOUT * 10))
                _st_label_608 -= 1

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._Node_ReceivedEvent_0 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_Node_ReceivedEvent_0', PatternExpr_789, sources=[PatternExpr_794], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def run(self):
        nacceptors = (int(sys.argv[1]) if (len(sys.argv) > 1) else 3)
        nproposers = (int(sys.argv[2]) if (len(sys.argv) > 2) else 5)
        nlearners = (int(sys.argv[3]) if (len(sys.argv) > 3) else 3)
        acceptors = self.new(Acceptor, num=nacceptors)
        proposers = self.new(Proposer, (acceptors,), num=nproposers)
        learners = self.new(Learner, (acceptors,), num=nlearners)
        for p in acceptors:
            self._setup(p, (learners,))
        self._start(((acceptors | proposers) | learners))
        super()._label('_st_label_780', block=False)
        l = None

        def UniversalOpExpr_781():
            nonlocal l
            for l in learners:
                if (not PatternExpr_796.match_iter(self._Node_ReceivedEvent_0, _BoundPattern802_=l)):
                    return False
            return True
        _st_label_780 = 0
        while (_st_label_780 == 0):
            _st_label_780 += 1
            if UniversalOpExpr_781():
                _st_label_780 += 1
            else:
                super()._label('_st_label_780', block=True)
                _st_label_780 -= 1
        self.output('done')
        self.send(('done',), to=(acceptors | proposers))



# -*- generated by 1.0.12 -*-
import da
PatternExpr_192 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern196_')])
PatternExpr_199 = da.pat.BoundPattern('_BoundPattern200_')
PatternExpr_224 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('abort'), da.pat.BoundPattern('_BoundPattern229_')])
PatternExpr_232 = da.pat.BoundPattern('_BoundPattern233_')
PatternExpr_254 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready'), da.pat.BoundPattern('_BoundPattern259_')])
PatternExpr_262 = da.pat.FreePattern('c')
PatternExpr_299 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready'), da.pat.BoundPattern('_BoundPattern304_')])
PatternExpr_307 = da.pat.BoundPattern('_BoundPattern308_')
PatternExpr_336 = da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern339_')])
PatternExpr_342 = da.pat.BoundPattern('_BoundPattern343_')
PatternExpr_234 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern240_')]), da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('abort'), da.pat.BoundPattern('_BoundPattern246_')])])
PatternExpr_309 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern315_')]), da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready'), da.pat.BoundPattern('_BoundPattern321_')])])
PatternExpr_344 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern350_')]), da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern354_')])])
PatternExpr_406 = da.pat.TuplePattern([da.pat.ConstantPattern('prepare'), da.pat.FreePattern('tid')])
PatternExpr_413 = da.pat.FreePattern('coord')
PatternExpr_444 = da.pat.TuplePattern([da.pat.ConstantPattern('abort'), da.pat.FreePattern('tid')])
PatternExpr_457 = da.pat.TuplePattern([da.pat.ConstantPattern('commit'), da.pat.FreePattern('tid')])
PatternExpr_464 = da.pat.FreePattern('coord')
PatternExpr_519 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_524 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
_config_object = {}
import sys
from random import randint

class Coordinator(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._CoordinatorReceivedEvent_0 = []
        self._CoordinatorReceivedEvent_1 = []
        self._CoordinatorReceivedEvent_2 = []
        self._CoordinatorReceivedEvent_3 = []
        self._CoordinatorReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_0', PatternExpr_192, sources=[PatternExpr_199], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_1', PatternExpr_224, sources=[PatternExpr_232], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_2', PatternExpr_254, sources=[PatternExpr_262], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_3', PatternExpr_299, sources=[PatternExpr_307], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_4', PatternExpr_336, sources=[PatternExpr_342], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, tid, cohorts, **rest_593):
        super().setup(tid=tid, cohorts=cohorts, **rest_593)
        self._state.tid = tid
        self._state.cohorts = cohorts
        pass

    def run(self):
        self.to_commit()
        self.send(('done',), to=self._state.cohorts)
        self.output('terminating')

    def to_commit(self):
        self.send(('prepare', self._state.tid), to=self._state.cohorts)
        super()._label('_st_label_183', block=False)
        c = None

        def UniversalOpExpr_184():
            nonlocal c
            for c in self._state.cohorts:

                def ExistentialOpExpr_190(c):
                    for (_, (_, _, _BoundPattern207_), (_ConstantPattern209_, _, _BoundPattern212_)) in self._CoordinatorReceivedEvent_0:
                        if (_BoundPattern207_ == c):
                            if (_ConstantPattern209_ == 'vote'):
                                if (_BoundPattern212_ == self._state.tid):
                                    if True:
                                        return True
                    return False
                if (not ExistentialOpExpr_190(c=c)):
                    return False
            return True
        _st_label_183 = 0
        while (_st_label_183 == 0):
            _st_label_183 += 1
            if UniversalOpExpr_184():
                _st_label_183 += 1
            else:
                super()._label('_st_label_183', block=True)
                _st_label_183 -= 1
        c = None

        def ExistentialOpExpr_217():
            nonlocal c
            for c in self._state.cohorts:
                if PatternExpr_234.match_iter(self._CoordinatorReceivedEvent_1, _BoundPattern240_=c, _BoundPattern246_=self._state.tid, SELF_ID=self._id):
                    return True
            return False
        if ExistentialOpExpr_217():
            s = {c for c in self._state.cohorts for (_, (_, _, _FreePattern270_), (_ConstantPattern272_, _ConstantPattern274_, _BoundPattern276_)) in self._CoordinatorReceivedEvent_2 if (_FreePattern270_ == c) if (_ConstantPattern272_ == 'vote') if (_ConstantPattern274_ == 'ready') if (_BoundPattern276_ == self._state.tid)}
            self.send(('abort', self._state.tid), to=s)
            self.abort(self._state.tid)
        c = None

        def UniversalOpExpr_292():
            nonlocal c
            for c in self._state.cohorts:
                if (not PatternExpr_309.match_iter(self._CoordinatorReceivedEvent_3, _BoundPattern315_=c, _BoundPattern321_=self._state.tid, SELF_ID=self._id)):
                    return False
            return True
        if UniversalOpExpr_292():
            self.send(('commit', self._state.tid), to=self._state.cohorts)
            super()._label('_st_label_328', block=False)
            c = None

            def UniversalOpExpr_329():
                nonlocal c
                for c in self._state.cohorts:
                    if (not PatternExpr_344.match_iter(self._CoordinatorReceivedEvent_4, _BoundPattern350_=c, _BoundPattern354_=self._state.tid, SELF_ID=self._id)):
                        return False
                return True
            _st_label_328 = 0
            while (_st_label_328 == 0):
                _st_label_328 += 1
                if UniversalOpExpr_329():
                    _st_label_328 += 1
                else:
                    super()._label('_st_label_328', block=True)
                    _st_label_328 -= 1
            self.commit(self._state.tid)

    def abort(self, tid):
        self.output(('abort:' + str(tid)))

    def commit(self, tid):
        self.output(('commit:' + str(tid)))

class Cohort(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._CohortReceivedEvent_3 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_0', PatternExpr_406, sources=[PatternExpr_413], destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_405]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_1', PatternExpr_444, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_443]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_2', PatternExpr_457, sources=[PatternExpr_464], destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_456]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_3', PatternExpr_519, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, failure_rate, **rest_593):
        super().setup(failure_rate=failure_rate, **rest_593)
        self._state.failure_rate = failure_rate
        pass

    def run(self):
        super()._label('_st_label_516', block=False)
        _st_label_516 = 0
        while (_st_label_516 == 0):
            _st_label_516 += 1
            if PatternExpr_524.match_iter(self._CohortReceivedEvent_3, SELF_ID=self._id):
                _st_label_516 += 1
            else:
                super()._label('_st_label_516', block=True)
                _st_label_516 -= 1

    def prepared(self, tid):
        return (randint(0, 100) > self._state.failure_rate)

    def ready(self, tid):
        self.output(('ready:' + str(tid)))

    def abort(self, tid):
        self.output(('abort:' + str(tid)))

    def commit(self, tid):
        self.output(('commit:' + str(tid)))

    def _Cohort_handler_405(self, tid, coord):
        if self.prepared(tid):
            self.send(('vote', 'ready', tid), to=coord)
            self.ready(tid)
        else:
            self.send(('vote', 'abort', tid), to=coord)
            self.abort(tid)
    _Cohort_handler_405._labels = None
    _Cohort_handler_405._notlabels = None

    def _Cohort_handler_443(self, tid):
        self.abort(tid)
    _Cohort_handler_443._labels = None
    _Cohort_handler_443._notlabels = None

    def _Cohort_handler_456(self, tid, coord):
        self.send(('done', tid), to=coord)
        self.commit(tid)
    _Cohort_handler_456._labels = None
    _Cohort_handler_456._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        ncohorts = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        fail_rate = (int(sys.argv[2]) if (len(sys.argv) > 2) else 10)
        cohorts = self.new(Cohort, (fail_rate,), num=ncohorts)
        coordinators = self.new(Coordinator, ('001', cohorts), num=1)
        self._start((cohorts | coordinators))



# -*- generated by 1.0.12 -*-
import da
PatternExpr_188 = da.pat.TuplePattern([da.pat.ConstantPattern('request'), da.pat.FreePattern('c2'), da.pat.FreePattern('p')])
PatternExpr_213 = da.pat.TuplePattern([da.pat.ConstantPattern('release'), da.pat.BoundPattern('_BoundPattern216_'), da.pat.BoundPattern('_BoundPattern217_')])
PatternExpr_246 = da.pat.TuplePattern([da.pat.ConstantPattern('ack'), da.pat.FreePattern('c2'), da.pat.BoundPattern('_BoundPattern250_')])
PatternExpr_282 = da.pat.TuplePattern([da.pat.ConstantPattern('request'), da.pat.FreePattern(None), da.pat.FreePattern('p')])
PatternExpr_330 = da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern333_')])
PatternExpr_336 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern346_')])])
PatternExpr_220 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('release'), da.pat.BoundPattern('_BoundPattern230_'), da.pat.BoundPattern('_BoundPattern231_')])])
_config_object = {}
import sys

class P(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._PReceivedEvent_0 = []
        self._PReceivedEvent_1 = []
        self._PReceivedEvent_2 = []
        self._PReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_0', PatternExpr_188, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_1', PatternExpr_213, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_2', PatternExpr_246, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_3', PatternExpr_282, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_281]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_4', PatternExpr_330, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, s, nrequests, **rest_411):
        super().setup(s=s, nrequests=nrequests, **rest_411)
        self._state.s = s
        self._state.nrequests = nrequests
        pass

    def run(self):

        def task():
            self.output('in critical section')
        for i in range(self._state.nrequests):
            self.mutex(task)
        self.send(('done', self._id), to=self._state.s)
        super()._label('_st_label_321', block=False)
        p = None

        def UniversalOpExpr_322():
            nonlocal p
            for p in self._state.s:
                if (not PatternExpr_336.match_iter(self._PReceivedEvent_4, _BoundPattern346_=p, SELF_ID=self._id)):
                    return False
            return True
        _st_label_321 = 0
        while (_st_label_321 == 0):
            _st_label_321 += 1
            if UniversalOpExpr_322():
                _st_label_321 += 1
            else:
                super()._label('_st_label_321', block=True)
                _st_label_321 -= 1
        self.output('terminating')

    def mutex(self, task):
        super()._label('request', block=False)
        c = self.logical_clock()
        self.send(('request', c, self._id), to=self._state.s)
        super()._label('_st_label_184', block=False)
        c2 = p = None

        def UniversalOpExpr_186():
            nonlocal c2, p
            for (_, _, (_ConstantPattern205_, c2, p)) in self._PReceivedEvent_0:
                if (_ConstantPattern205_ == 'request'):
                    if (not (PatternExpr_220.match_iter(self._PReceivedEvent_1, _BoundPattern230_=c2, _BoundPattern231_=p, SELF_ID=self._id) or ((c, self._id) < (c2, p)))):
                        return False
            return True
        c2 = p = None

        def UniversalOpExpr_239():
            nonlocal c2, p
            for p in self._state.s:

                def ExistentialOpExpr_244(p):
                    nonlocal c2
                    for (_, _, (_ConstantPattern261_, c2, _BoundPattern264_)) in self._PReceivedEvent_2:
                        if (_ConstantPattern261_ == 'ack'):
                            if (_BoundPattern264_ == p):
                                if (c2 > c):
                                    return True
                    return False
                if (not ExistentialOpExpr_244(p=p)):
                    return False
            return True
        _st_label_184 = 0
        while (_st_label_184 == 0):
            _st_label_184 += 1
            if (UniversalOpExpr_186() and UniversalOpExpr_239()):
                _st_label_184 += 1
            else:
                super()._label('_st_label_184', block=True)
                _st_label_184 -= 1
        super()._label('critical_section', block=False)
        task()
        super()._label('release', block=False)
        self.send(('release', c, self._id), to=self._state.s)

    def _P_handler_281(self, p):
        self.send(('ack', self.logical_clock(), self._id), to=p)
    _P_handler_281._labels = None
    _P_handler_281._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])
    _config_object = {'channel': 'Fifo', 'clock': 'Lamport'}

    def run(self):
        nprocs = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        nrequests = (int(sys.argv[2]) if (len(sys.argv) > 2) else 1)
        ps = self.new(P, num=nprocs)
        for p in ps:
            self._setup(p, ((ps - {p}), nrequests))
        self._start(ps)



# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from ast import *
from itertools import chain
from da.compiler import dast
from da.compiler.utils import printd, printw, printe

OperatorMap = {
    dast.AddOp : Add,
    dast.SubOp : Sub,
    dast.MultOp : Mult,
    dast.DivOp : Div,
    dast.ModOp : Mod,
    dast.PowOp : Pow,
    dast.LShiftOp : LShift,
    dast.RShiftOp : RShift,
    dast.BitOrOp : BitOr,
    dast.BitXorOp : BitXor,
    dast.BitAndOp : BitAnd,
    dast.FloorDivOp : FloorDiv,

    dast.EqOp : Eq,
    dast.NotEqOp : NotEq,
    dast.LtOp : Lt,
    dast.LtEOp : LtE,
    dast.GtOp : Gt,
    dast.GtEOp : GtE,
    dast.IsOp : Is,
    dast.IsNotOp : IsNot,
    dast.InOp : In,
    dast.NotInOp : NotIn,

    dast.USubOp : USub,
    dast.UAddOp : UAdd,
    dast.InvertOp : Invert,

    dast.AndOp : And,
    dast.OrOp : Or
}
# New matrix multiplication operator since 3.5:
if sys.version_info > (3, 5):
    OperatorMap[dast.MatMultOp] = MatMult

AggregateMap = {
    dast.MaxExpr : "max",
    dast.MinExpr : "min",
    dast.SizeExpr : "len",
    dast.SumExpr : "sum"
}

CONFIG_OBJECT_NAME = "_config_object"
STATE_ATTR_NAME = "_state"
ENTRYPOINT_NAME = "run"
CATCHALL_PARAM_NAME = "rest_%d"

########## Convenience methods for creating AST nodes: ##########

def call_noarg_ast(name):
    return Call(Name(name, Load()), [], [], None, None)

def pyCall(func, args=[], keywords=[], starargs=None, kwargs=None):
    if isinstance(func, str):
        func = pyName(func)
    if sys.version_info < (3, 5):
        ast = Call(func,
                   list(args),
                   [keyword(arg, val) for arg, val in keywords],
                   starargs,
                   kwargs)
        propagate_attributes((starargs, kwargs), ast)
    else:
        ast = Call(func,
                   list(args),
                   [keyword(arg, val) for arg, val in keywords])
    propagate_attributes(func, ast)
    propagate_attributes(args, ast)
    propagate_attributes([val for _, val in keywords], ast)
    return ast

def pyName(name, ctx=None):
    return Name(name, Load() if ctx is None else ctx)

def pyNone():
    if sys.version_info > (3, 4):
        return NameConstant(None)
    else:
        return pyName("None")

def pyTrue():
    if sys.version_info > (3, 4):
        return NameConstant(True)
    else:
        return pyName("True")

def pyFalse():
    if sys.version_info > (3, 4):
        return NameConstant(False)
    else:
        return pyName("False")

def pyNot(expr):
    ast = UnaryOp(Not(), expr)
    return propagate_attributes(expr, ast)

def pyList(elts, ctx=None):
    ast = List(elts, Load() if ctx is None else ctx)
    return propagate_attributes(elts, ast)

def pySet(elts, ctx=None):
    ast = Set(elts)
    return propagate_attributes(elts, ast)

def pyTuple(elts, ctx=None):
    ast = Tuple(elts, Load() if ctx is None else ctx)
    return propagate_attributes(elts, ast)

def pySetC(elts):
    return pyCall("set", args=elts)

def pySubscr(value, index, ctx=None):
    ast = Subscript(value, Index(index),
                    Load() if ctx is None else ctx)
    return propagate_attributes((value, index), ast)

def pySize(value):
    return pyCall("len", [value])

def pyMin(value):
    return pyCall("min", [value])

def pyMax(value):
    return pyCall("max", [value])

def pyAttr(name, attr, ctx=None):
    if isinstance(name, str):
        ast = Attribute(Name(name, Load()), attr,
                         Load() if ctx is None else ctx)
    else:
        ast = Attribute(name, attr, Load() if ctx is None else ctx)
    return propagate_attributes(ast.value, ast)

def pyCompare(left, op, right):
    ast = Compare(left, [op()], [right])
    return propagate_attributes((left, right), ast)

def pyLabel(name, block=False, timeout=None):
    kws = [("block", pyTrue() if block else pyFalse())]
    if timeout is not None:
        kws.append(("timeout", timeout))
    return Expr(pyCall(func=pyAttr(pyCall("super"), "_label"),
                       args=[Str(name)],
                       keywords=kws))

def pycomprehension(target, iter, ifs, is_async=0):
    if sys.version_info < (3, 6):
        return comprehension(target, iter, ifs)
    else:
        return comprehension(target, iter, ifs, is_async)

def pyClassDef(name, bases=[], keywords=[], starargs=None,
               kwargs=None, body=[], decorator_list=[]):
    if sys.version_info < (3, 5):
        return ClassDef(name,
                        list(bases),
                        [keyword(arg, val) for arg, val in keywords],
                        starargs,
                        kwargs,
                        list(body),
                        list(decorator_list))
    else:
        return ClassDef(name,
                        list(bases),
                        [keyword(arg, val) for arg, val in keywords],
                        list(body),
                        list(decorator_list))

def pyFunctionDef(name, args=[], body=[], decorator_list=[], returns=None):
    arglist = arguments(args=[arg(n, None) for n in args],
                        vararg=None,
                        varargannotation=None,
                        kwonlyargs=[],
                        kwarg=None,
                        kwargannotation=None,
                        defaults=[],
                        kw_defaults=None)
    return FunctionDef(name,
                       arglist,
                       list(body),
                       list(decorator_list),
                       returns)

def propagate_attributes(from_nodes, to_node):
    if isinstance(to_node, AST):
        if not (isinstance(from_nodes, list) or
                isinstance(from_nodes, tuple) or
                isinstance(from_nodes, set)):
            from_nodes = [from_nodes]
        for fro in from_nodes:
            if (hasattr(fro, "prebody") and isinstance(fro.prebody, list)):
                if not hasattr(to_node, "prebody"):
                    to_node.prebody = []
                to_node.prebody.extend(fro.prebody)
            if (hasattr(fro, "postbody") and isinstance(fro.postbody, list)):
                if not hasattr(to_node, "postbody"):
                    to_node.postbody = []
                to_node.postbody.extend(fro.postbody)
    return to_node

def propagate_fields(node):
    if hasattr(node, '_fields'):
        for f in node._fields:
            propagate_attributes(getattr(node, f), node)
    return node

def concat_bodies(subexprs, body):
    prebody = []
    postbody = []
    for e in subexprs:
        if hasattr(e, "prebody"):
            prebody.extend(e.prebody)
        if hasattr(e, "postbody"):
            postbody.extend(e.postbody)
    return prebody + body + postbody

def is_all_wildcards(targets):
    """True if 'targets' contain only wildcards."""

    for elt in targets:
        if not (isinstance(elt, Name) and elt.id == '_'):
            return False
    return True

class PythonGeneratorException(Exception): pass

def translate(distalgo_ast, filename="", options=None):
    pg = PythonGenerator(filename, options)
    try:
        return pg.visit(distalgo_ast)
    except Exception as ex:
        raise PythonGeneratorException(str(pg.current_node)) from ex

# List of arguments needed to initialize a process:
PROC_INITARGS = ["procimpl", "props"]

PREAMBLE = parse(
    """
import da
    """).body
POSTAMBLE = parse("""
if __name__ == "__main__":
    da.init(config)
""").body


class PythonGenerator(NodeVisitor):
    """Transforms DistPy AST into Python AST.

    """

    def __init__(self, filename="", options=None):
        self.filename = filename
        self.processed_patterns = set()
        self.preambles = list(PREAMBLE)
        self.postambles = list()
        # One instance of PatternComprehensionGenerator for each query.
        # This is needed so free vars with the same name in a query can be
        # properly unified:
        self.pattern_generator = None
        self.cmdline_args = options
        self.module_args = None
        # Used by incgen to avoid expanding 'pre/postbody' in the inc module:
        self.disable_body_expansion = False

        self.current_node = None

    def get_option(self, option, default=None):
        if hasattr(self.cmdline_args, option):
            return getattr(self.cmdline_args, option)
        elif hasattr(self.module_args, option):
            return getattr(self.module_args, option)
        else:
            return default

    def reset(self):
        """Resets internal states.

        Call this before compiling a new file or code segment, if you don't
        want to create a new instance.

        """
        self.processed_patterns = set()
        self.preambles = list(PREAMBLE)
        self.postambles = list()
        self.pattern_generator = None

    def visit(self, node):
        """Generic visit method.

        If the Incrementalization interface generated code for this node, as
        indicated by the 'ast_override' attribute, then return the generated
        code. Otherwise, call the normal visit method.

        """
        if node is None:
            return None

        assert isinstance(node, dast.DistNode)
        self.current_node = node
        if hasattr(node, "ast_override"):
            res = node.ast_override
        else:
            res = super().visit(node)

        if isinstance(res, list) and not self.disable_body_expansion:
            # This is a statement, expand pre and post bodies:
            return concat_bodies([node], res)
        else:
            # This is an expression, pass on pre and post bodies:
            return propagate_attributes([node], res)

    def body(self, body, res=None):
        """Process a block of statements."""
        if res is None:
            res = []
        for stmt in body:
            if stmt.label is not None:
                res.append(pyLabel(stmt.label))
            ast = self.visit(stmt)
            if ast is not None:
                res.extend(ast)
            else:
                printe("None result from %s" % str(stmt))
        return res

    def bases(self, bases):
        """Process base classes of a class definition."""
        res = []
        for expr in bases:
            res.append(self.visit(expr))
        return res

    def visit_Program(self, node):
        self.module_args = node._compiler_options
        mainbody = self.body(node.body)
        if node.nodecls is not None:
            # `nodecls` is the `Node_` process:
            nodeproc = self.visit(node.nodecls)
        body = list(self.preambles)
        body.append(self.generate_config(node))
        body.extend(mainbody)
        if node.nodecls is not None:
            body.extend(nodeproc)
        body.extend(self.postambles)
        return Module(body)

    def generate_config(self, node):
        return Assign([pyName(CONFIG_OBJECT_NAME)],
                      Dict([Str(key) for key, _ in node.configurations],
                           [self.visit(val) for _, val in node.configurations]))

    def generate_event_def(self, node):
        evtype = pyAttr(pyAttr("da", "pat"), node.type.__name__)
        name = Str(node.name)
        history = self.history_stub(node)
        pattern = self.visit(node.pattern)
        sources = pyNone()
        destinations = pyNone()
        timestamps = pyNone()
        if len(node.sources) > 0:
            sources = pyList([self.visit(s) for s in node.sources])
        if len(node.destinations) > 0:
            destinations = pyList([self.visit(s) for s in node.destinations])
        if len(node.timestamps) > 0:
            timestamps = pyList([self.visit(s) for s in node.timestamps])
        handlers = pyList([pyAttr("self", h.name) for h in node.handlers])
        return pyCall(func=pyAttr(pyAttr("da", "pat"), "EventPattern"),
                      args=[evtype, name, pattern],
                      keywords=[("sources", sources),
                                ("destinations", destinations),
                                ("timestamps", timestamps),
                                ("record_history", history),
                                ("handlers", handlers)])

    def history_initializers(self, node):
        return [Assign(targets=[pyAttr("self", evt.name)],
                       value=pyList([]))
                for evt in node.events if evt.record_history]

    def generate_init(self, node):
        supercall = [Expr(pyCall(func=pyAttr(pyCall(pyName("super")),
                                             "__init__"),
                                 args=[pyName(n) for n in PROC_INITARGS]))]
        histories = self.history_initializers(node)
        events = [Expr(pyCall(func=pyAttr(pyAttr("self", "_events"), "extend"),
                              args=[pyList([self.generate_event_def(evt)
                              for evt in node.events])]))]
        return pyFunctionDef(name="__init__",
                             args=(["self"] + PROC_INITARGS),
                             body=(supercall + histories + events))

    def generate_handlers(self, node):
        """Generate the message handlers of a process."""
        body = []
        for evt in node.events:
            for handler in evt.handlers:
                body.extend((self.visit(handler)))
        return body

    def visit_Arguments(self, node):
        """Generates the argument lists for functions and lambdas."""
        args = [arg(ident.name, None) for ident in node.args]
        kwonlyargs = [arg(ident.name, None) for ident in node.kwonlyargs]
        kw_defaults = [self.visit(expr) for expr in node.kw_defaults]
        defaults = [self.visit(expr) for expr in node.defaults]
        if sys.version_info >= (3, 5):
            vararg = arg(node.vararg.name, None) \
                     if node.vararg is not None else None
            kwarg = arg(node.kwarg.name, None) \
                    if node.kwarg is not None else None
            return arguments(
                args=args,
                vararg=vararg,
                kwonlyargs=kwonlyargs,
                kwarg=kwarg,
                defaults=defaults,
                kw_defaults=kw_defaults)
        else:
            vararg = node.vararg.name if node.vararg is not None else None
            kwarg = node.kwarg.name if node.kwarg is not None else None
            return arguments(
                args=args,
                vararg=vararg,
                varargannotation=None,
                kwonlyargs=kwonlyargs,
                kwarg=kwarg,
                kwargannotation=None,
                defaults=defaults,
                kw_defaults=kw_defaults)

    def visit_Process(self, node):
        printd("Compiling process %s" % node.name)
        printd("has methods:%r" % node.methods)
        cd = ClassDef()
        cd.name = node.name
        cd.bases = self.bases(node.bases)
        if node is node.immediate_container_of_type(dast.Program).nodecls:
            cd.bases.append(pyAttr("da", "NodeProcess"))
        else:
            cd.bases.append(pyAttr("da", "DistProcess"))
        if node.ast is not None:
            # ########################################
            # TODO: just pass these through until we figure out a use for them:
            cd.keywords = node.ast.keywords
            if sys.version_info < (3, 5):
                cd.starargs = node.ast.starargs
                cd.kwargs = node.ast.kwargs
            # ########################################
        else:
            cd.keywords = []
            cd.starargs = []
            cd.kwargs = []
        cd.body = [self.generate_init(node)]
        if node.configurations:
            cd.body.append(self.generate_config(node))
        if node.setup is not None:
            cd.body.extend(self.visit(node.setup))
        if node.entry_point is not None:
            cd.body.extend(self._entry_point(node.entry_point))
        cd.decorator_list = [self.visit(d) for d in node.decorators]
        cd.body.extend(self.body(node.methods))
        cd.body.extend(self.generate_handlers(node))
        return [cd]

    def _entry_point(self, node):
        stmts = self.visit(node)
        stmts[0].name = ENTRYPOINT_NAME
        stmts[0].args = [arg('self', None)]
        return stmts

    def _generate_setup(self, node, fd):
        fd.args = self.visit(node.parent.args)
        kwargname = CATCHALL_PARAM_NAME % node._index
        fd.args.kwarg = arg(kwargname, None)
        superargs = [(argname.name, pyName(argname.name))
                     for argname in node.parent.args.args]
        superargs.append((None, pyName(kwargname)))
        fd.body.append(Expr(
            pyCall(func=pyAttr(pyCall("super"), "setup"),
                   keywords=superargs)))
        fd.body.extend([
            Assign(targets=[pyAttr(pyAttr("self", STATE_ATTR_NAME),
                                   name, Store())],
                   value=pyName(name))
            for name in node.parent.ordered_names
        ])

    def visit_Function(self, node):
        fd = FunctionDef()
        fd.name = node.name
        fd.args = self.visit(node.args)
        fd.body = []
        if isinstance(node.parent, dast.Process):
            if node.name == "setup":
                self._generate_setup(node, fd)
            fd.args.args.insert(0, arg("self", None))
        fd.body = self.body(node.body, fd.body)
        fd.decorator_list = [self.visit(d) for d in node.decorators]
        fd.returns = None
        return [fd]

    def visit_ClassStmt(self, node):
        cd = pyClassDef(name=node.name,
                        bases=self.bases(node.bases),
                        body=self.body(node.body))
        # ########################################
        # TODO: just pass these through until we figure out a use for them:
        cd.keywords = node.ast.keywords
        if sys.version_info < (3, 5):
            cd.starargs = node.ast.starargs
            cd.kwargs = node.ast.kwargs
        # ########################################
        cd.decorator_list = [self.visit(d) for d in node.decorators]
        return [cd]

    def visit_PythonExpr(self, node):
        return node.ast

    def visit_SimpleExpr(self, node):
        return self.visit(node.value)

    def visit_AttributeExpr(self, node):
        sub = self.visit(node.value)
        ast = Attribute(sub, node.attr, None)
        return propagate_attributes([sub], ast)

    def visit_SubscriptExpr(self, node):
        val = self.visit(node.value)
        if isinstance(node.index, dast.SliceExpr):
            idx = self.visit(node.index)
        else:
            idx = Index(self.visit(node.index))
            propagate_attributes([idx.value], idx)
        ast = Subscript(val, idx, Load())
        return propagate_attributes((val, idx), ast)

    def visit_SliceExpr(self, node):
        l = self.visit(node.lower) if node.lower is not None else None
        u = self.visit(node.upper) if node.upper is not None else None
        s = self.visit(node.step) if node.step is not None else None
        ast = Slice(l, u, s)
        return propagate_attributes((l, u, s), ast)

    def visit_StarredExpr(self, node):
        val = self.visit(node.value)
        ast = Starred(val, None)
        return propagate_attributes([val], ast)

    def visit_EllipsisExpr(self, node):
        return Ellipsis()

    def visit_ConstantExpr(self, node):
        if isinstance(node.value, str):
            return Str(node.value)
        elif isinstance(node.value, bytes):
            return Bytes(node.value)
        else:
            return Num(node.value)

    def visit_SelfExpr(self, node):
        return pyAttr("self", "_id")

    def visit_TrueExpr(self, node):
        return pyTrue()

    def visit_FalseExpr(self, node):
        return pyFalse()

    def visit_NoneExpr(self, node):
        return pyNone()

    def visit_FormattedValueExpr(self, node):
        return FormattedValue(self.visit(node.value), node.conversion,
                              self.visit(node.format_spec)
                              if node.format_spec else None)

    def visit_TupleExpr(self, node):
        return pyTuple([self.visit(e) for e in node.subexprs])

    def visit_ListExpr(self, node):
        return pyList([self.visit(e) for e in node.subexprs])

    def visit_SetExpr(self, node):
        return pySet([self.visit(e) for e in node.subexprs])

    def visit_FormattedStrExpr(self, node):
        return JoinedStr([self.visit(value) for value in node.subexprs])

    def visit_DictExpr(self, node):
        ast = Dict([self.visit(e) for e in node.keys],
                   [self.visit(e) for e in node.values])
        return propagate_attributes(ast.keys + ast.values, ast)

    def visit_IfExpr(self, node):
        ast = IfExp(self.visit(node.condition),
                    self.visit(node.body),
                    self.visit(node.orbody))
        return propagate_attributes((ast.test, ast.body, ast.orelse), ast)

    def visit_CallExpr(self, node):
        return pyCall(self.visit(node.func),
                     [self.visit(a) for a in node.args],
                     [(key, self.visit(value)) for key, value in node.keywords],
                     self.visit(node.starargs)
                     if node.starargs is not None else None,
                     self.visit(node.kwargs)
                     if node.kwargs is not None else None)

    def visit_ApiCallExpr(self, node):
        return pyCall(pyAttr("da", node.func),
                     [self.visit(a) for a in node.args],
                     [(key, self.visit(value)) for key, value in node.keywords],
                     self.visit(node.starargs)
                     if node.starargs is not None else None,
                     self.visit(node.kwargs)
                     if node.kwargs is not None else None)

    def visit_BuiltinCallExpr(self, node):
        return pyCall(pyAttr("self", node.func),
                     [self.visit(a) for a in node.args],
                     [(key, self.visit(value)) for key, value in node.keywords],
                     self.visit(node.starargs)
                     if node.starargs is not None else None,
                     self.visit(node.kwargs)
                     if node.kwargs is not None else None)

    visit_SetupExpr = visit_StartExpr = visit_ConfigExpr = visit_BuiltinCallExpr

    def visit_AggregateExpr(self, node):
        return pyCall(AggregateMap[type(node)],
                      [self.visit(a) for a in node.args],
                      [(key, self.visit(value)) for key, value in node.keywords],
                      self.visit(node.starargs)
                      if node.starargs is not None else None,
                      self.visit(node.kwargs)
                      if node.kwargs is not None else None)

    visit_MaxExpr = visit_AggregateExpr
    visit_MinExpr = visit_AggregateExpr
    visit_SumExpr = visit_AggregateExpr
    visit_SizeExpr = visit_AggregateExpr

    def visit_LogicalExpr(self, node):
        if node.operator is dast.NotOp:
            ast = UnaryOp(Not(), self.visit(node.left))
            return propagate_attributes([ast.operand], ast)
        else:
            ast = BoolOp(OperatorMap[node.operator](),
                         [self.visit(e) for e in node.subexprs])
            return propagate_attributes(ast.values, ast)

    def visit_DomainSpec(self, node):
        domain = self.visit(node.domain)
        if not isinstance(node.pattern, dast.PatternExpr):
            result = pycomprehension(self.visit(node.pattern), domain, [])
        else:
            if self.pattern_generator is None:
                # Legacy pattern
                target, condlist = PatternComprehensionGenerator().visit(
                    node.pattern)
            else:
                target, condlist = self.pattern_generator.visit(node.pattern)
            result = pycomprehension(target, domain, condlist)
        return propagate_fields(result)

    def visit_QuantifiedExpr(self, node):
        if self.pattern_generator is None:
            self.pattern_generator = PatternComprehensionGenerator()
            is_top_level_query = True
        else:
            is_top_level_query = False
            if not self.get_option('use_top_semantic', default=False):
                self.pattern_generator.push_state()
                self.pattern_generator.reset_state()

        body = funcbody = []
        for domspec in node.domains:
            comp = self.visit(domspec)
            ast = For(comp.target, comp.iter, [], [])
            body.append(propagate_attributes([ast.iter], ast))
            body = body[0].body
            for cond in comp.ifs:
                ast = If(cond, [], [])
                body.append(propagate_attributes([ast.test], ast))
                body = body[0].body
        postbody = []
        ifcond = self.visit(node.predicate)
        if hasattr(ifcond, "prebody"):
            body.extend(ifcond.prebody)
        if hasattr(ifcond, "postbody"):
            postbody.extend(cnode.postbody)

        if node.operator is dast.UniversalOp:
            ifcond = UnaryOp(Not(), ifcond)
            ifbody = [Return(pyFalse())]
        else:                   # ExistentialExpr
            ifbody = [Return(pyTrue())]
        body.append(If(ifcond, ifbody, []))
        body.extend(postbody)
        if node.operator is dast.UniversalOp:
            funcbody.append(Return(pyTrue()))
        else:
            funcbody.append(Return(pyFalse()))

        # names that should be unified with a containing query need to be
        # explicitly passed in:
        curnode = node
        params = set()
        while curnode is not node.top_level_query:
            curnode = curnode.parent
            if isinstance(curnode, dast.QueryExpr):
                params |= set(curnode.ordered_local_freevars)
        params &= node.nameobjs
        ast = pyCall(func=pyName(node.name),
                     keywords=[(v.name, self.visit(v)) for v in params])
        funast = pyFunctionDef(name=node.name,
                               args=[v.name for v in params],
                               body=funcbody)
        ast.prebody = [funast]

        nameset = node.freevars - params
        if len(nameset) > 0:
            # Back patch nonlocal statement
            if not isinstance(node.scope, dast.ComprehensionExpr):
                if not isinstance(node.statement.parent, dast.Program):
                    decl = Nonlocal([nv.name for nv in nameset])
                else:
                    decl = Global([nv.name for nv in nameset])
                funast.body.insert(0, decl)

            # Assignment needed to ensure all vars are bound at this point
            if is_top_level_query:
                ast.prebody.insert(
                    0, Assign(targets=[pyName(nv.name) for nv in nameset],
                              value=pyNone()))

        if is_top_level_query:
            self.pattern_generator = None
        elif not self.get_option('use_top_semantic', default=False):
            self.pattern_generator.pop_state()
        return ast

    def visit_ComprehensionExpr(self, node):
        printd("Entering comprehension " + str(node))
        if self.pattern_generator is None:
            self.pattern_generator = PatternComprehensionGenerator()
            is_top_level_query = True
        else:
            self.pattern_generator.push_state()
            is_top_level_query = False
            if not self.get_option('use_top_semantic', default=False):
                self.pattern_generator.reset_state()

        generators = []
        dangling = []
        for dom in node.conditions:
            comp = self.visit(dom)
            if isinstance(comp, comprehension):
                # Tuck any dangling conditions here:
                comp.ifs.extend(dangling)
                dangling = []
                generators.append(comp)
            else:
                if len(generators) > 0:
                    generators[-1].ifs.append(comp)
                    propagate_attributes(generators[-1].ifs, generators[-1])
                else:
                    dangling.append(comp)
        if len(dangling) == 0:
            test = pyTrue()
        elif len(dangling) == 1:
            test = dangling[0]
        else:
            test = propagate_fields(BoolOp(And(), dangling))
        try:
            if type(node) is dast.DictCompExpr:
                key = self.visit(node.elem.key)
                value = self.visit(node.elem.value)
                if len(generators) > 0:
                    ast = DictComp(key, value, generators)
                else:
                    # No generators, degenerate to IfExp:
                    ast = IfExp(test,
                                propagate_fields(Dict([key], [value])),
                                Dict([], []))
                return propagate_fields(ast)
            else:
                elem = self.visit(node.elem)
                if len(generators) > 0:
                    if isinstance(node, dast.SetCompExpr):
                        ast = SetComp(elem, generators)
                    elif isinstance(node, dast.ListCompExpr):
                        ast = ListComp(elem, generators)
                    elif isinstance(node, dast.TupleCompExpr):
                        ast = pyCall("tuple", args=[GeneratorExp(elem, generators)])
                    elif isinstance(node, dast.GeneratorExpr):
                        ast = GeneratorExp(elem, generators)
                    else:
                        self.error("Unknown expression", node)
                        return None
                else:
                    # No generators, degenerate to IfExp:
                    if isinstance(node, dast.SetCompExpr):
                        ast = IfExp(test,
                                    propagate_fields(pySet([elem])),
                                    pySetC([]))
                    elif isinstance(node, dast.ListCompExpr):
                        ast = IfExp(test,
                                    propagate_fields(pyList([elem])),
                                    pyList([]))
                    elif isinstance(node, dast.TupleCompExpr):
                        ast = IfExp(test,
                                    propagate_fields(pyTuple([elem])),
                                    pyTuple([]))
                    elif isinstance(node, dast.GeneratorExpr):
                        # Impossible:
                        self.error("Illegal generator expression.", node)
                        return None
                    else:
                        self.error("Illegal unknown expression.", node)
                        return None
                return propagate_fields(ast)

        finally:
            if is_top_level_query:
                printd("Leaving toplevel " + str(node))
                self.pattern_generator = None
            else:
                # We need to restore the pattern state because comprehensions
                # does not bind witness values outside its scope:
                self.pattern_generator.pop_state()
                printd("Leaving comprehension " + str(node))

    visit_GeneratorExpr = visit_ComprehensionExpr
    visit_SetCompExpr = visit_ComprehensionExpr
    visit_ListCompExpr = visit_ComprehensionExpr
    visit_DictCompExpr = visit_ComprehensionExpr
    visit_TupleCompExpr = visit_ComprehensionExpr

    def visit_ComparisonExpr(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        if isinstance(node.left, dast.PatternExpr):
            # 'PATTERN in DOMAIN'
            context = [(v.unique_name, self.visit(v.value))
                       for v in node.left.ordered_boundpatterns]
            if node.immediate_container_of_type(dast.Process) is not None:
                # Propagate value of current process id to `SelfPattern`:
                context.append(("SELF_ID", pyAttr("self", "_id")))
            ast = pyCall(func=pyAttr(left, "match_iter"),
                         args=[right], keywords=context)
        else:
            op = OperatorMap[node.comparator]()
            ast = Compare(left, [op], [right])
        return propagate_attributes((left, right), ast)

    def visit_ArithmeticExpr(self, node):
        op = OperatorMap[node.operator]()
        if issubclass(node.operator, dast.UnaryOperator):
            ast = UnaryOp(op, self.visit(node.right))
            return propagate_fields(ast)
        else:
            ast = BinOp(self.visit(node.left), op, self.visit(node.right))
            return propagate_fields(ast)

    visit_BinaryExpr = visit_ArithmeticExpr
    visit_UnaryExpr = visit_ArithmeticExpr

    def visit_PatternElement(self, node):
        if type(node) is dast.FreePattern:
            val = Str(node.value.name) if node.value is not None else pyNone()
        elif type(node) is dast.BoundPattern:
            val = Str(node.unique_name)
        elif type(node) is dast.ConstantPattern:
            if isinstance(node.value, dast.SelfExpr):
                # We have to special case the 'self' expr here:
                return pyCall(func=pyAttr(pyAttr("da", "pat"), "SelfPattern"))
            else:
                val = self.visit(node.value)
        else:
            val = pyList([self.visit(v) for v in node.value])

        return pyCall(func=pyAttr(pyAttr("da", "pat"), type(node).__name__),
                      args=[val])

    visit_FreePattern = visit_PatternElement
    visit_BoundPattern = visit_PatternElement
    visit_ConstantPattern = visit_PatternElement
    visit_TuplePattern = visit_PatternElement
    visit_ListPattern = visit_PatternElement

    def visit_PatternExpr(self, node):
        if node.name not in self.processed_patterns:
            patast = self.visit(node.pattern)
            ast = Assign([pyName(node.name)], patast)
            self.preambles.append(ast)
            self.processed_patterns.add(node.name)
        return pyName(node.name)

    visit_LiteralPatternExpr = visit_PatternExpr

    def visit_HistoryExpr(self, node):
        assert node.event is not None
        return pyAttr("self", node.event.name)

    visit_ReceivedExpr = visit_HistoryExpr
    visit_SentExpr = visit_HistoryExpr

    def visit_LambdaExpr(self, node):
        args = self.visit(node.args)
        return Lambda(args, self.visit(node.body))

    def visit_NamedVar(self, node):
        if isinstance(node.scope, dast.Process):
            if node.name in node.scope.methodnames:
                return pyAttr("self", node.name)
            else:
                return pyAttr(pyAttr("self", STATE_ATTR_NAME), node.name)
        else:
            return pyName(node.name)


    ########## Statements ##########

    def visit_NoopStmt(self, node):
        return [Pass()]

    def visit_PassStmt(self, node):
        return [Pass()]

    def visit_AssignmentStmt(self, node):
        if node.value is None:
            # This is a "pure" annotation (since Python 3.6), don't generate
            # anything:
            return []
        targets = [self.visit(tgt) for tgt in node.targets]
        val = self.visit(node.value)
        ast = Assign(targets, val)
        return concat_bodies(targets + [val], [ast])

    def visit_OpAssignmentStmt(self, node):
        target = self.visit(node.target)
        val = self.visit(node.value)
        ast = AugAssign(target, OperatorMap[node.operator](), val)
        return concat_bodies([target, val], [ast])

    def visit_IfStmt(self, node):
        test = self.visit(node.condition)
        body = self.body(node.body)
        orelse = self.body(node.elsebody)
        ast = If(test, body, orelse)
        return concat_bodies([test], [ast])

    def visit_WhileStmt(self, node):
        test = self.visit(node.condition)
        body = self.body(node.body)
        orelse = self.body(node.elsebody)
        ast = While(test, body, orelse)
        return concat_bodies([test], [ast])

    def visit_ForStmt(self, node):
        comp = self.visit(node.domain)
        body = self.body(node.body)
        orelse = self.body(node.elsebody)
        ast = For(comp.target, comp.iter, body, orelse)
        return concat_bodies((comp.target, comp.iter), [ast])

    def visit_TryStmt(self, node):
        body = self.body(node.body)
        handlers = [self.visit(eh) for eh in node.excepthandlers]
        orelse = self.body(node.elsebody)
        finalbody = self.body(node.finalbody)
        return [Try(body, handlers, orelse, finalbody)]

    def visit_ExceptHandler(self, node):
        type = self.visit(node.type)
        body = self.body(node.body)
        return ExceptHandler(type, node.name, body)

    def visit_TryFinallyStmt(self, node):
        body = self.body(node.body)
        finalbody = self.body(node.finalbody)
        return [TryFinally(body, finalbody)]

    def visit_AwaitStmt(self, node):
        INCGRD = AugAssign(pyName(node.unique_label), Add(), Num(1))
        DEDGRD = AugAssign(pyName(node.unique_label), Sub(), Num(1))
        conds = []
        body = [INCGRD]
        labelname = node.label
        label = pyLabel(labelname, block=True,
                        timeout=(self.visit(node.timeout) if node.timeout is
                                 not None else None))
        last = body
        for br in node.branches:
            cond = self.visit(br.condition)
            conds.append(cond)
            ifbody = self.body(br.body)
            ifbody.append(INCGRD)
            brnode = If(cond, ifbody, [])
            last.append(brnode)
            last = brnode.orelse
        if node.timeout is not None:
            cond = pyAttr("self", "_timer_expired")
            ifbody = self.body(node.orelse)
            ifbody.append(INCGRD)
            brnode = If(cond, ifbody, [])
            last.append(brnode)
            last = brnode.orelse
        # Label call must come after the If tests:
        last.append(label)
        last.append(DEDGRD)
        whilenode = While(pyCompare(pyName(node.unique_label), Eq, Num(0)),
                          body, [])
        main = [Assign([pyName(node.unique_label)], Num(0))]
        if node.timeout is not None:
            main.append(Expr(pyCall(pyAttr("self", "_timer_start"))))
        main.append(whilenode)
        if node.is_in_loop:
            whilenode.orelse = [If(pyCompare(pyName(node.unique_label),
                                             NotEq, Num(2)),
                                   [Continue()], [])]
            main.append(If(pyCompare(pyName(node.unique_label),
                                             NotEq, Num(2)),
                                   [Break()], []))
        return concat_bodies(conds, main)

    def visit_LoopingAwaitStmt(self, node):
        INCGRD = AugAssign(pyName(node.unique_label), Add(), Num(1))
        labelname = node.label
        body = [If(pyCompare(pyName(node.unique_label), Eq, Num(2)),
                   [Break()],
                   [(If(pyCompare(pyName(node.unique_label), Eq, Num(1)),
                        [pyLabel(labelname, block=True,
                                 timeout=(self.visit(node.timeout)
                                          if node.timeout is not None
                                          else None))], []))]),
                Assign([pyName(node.unique_label)], Num(1))]
        conds = []
        last = body
        for br in node.branches:
            cond = self.visit(br.condition)
            conds.append(cond)
            ifbody = self.body(br.body)
            brnode = If(cond, ifbody, [])
            last.append(brnode)
            last = brnode.orelse
        if node.timeout is not None:
            cond = pyAttr("self", "_timer_expired")
            ifbody = self.body(node.orelse)
            ifbody.append(INCGRD)
            brnode = If(cond, ifbody, [])
            last.append(brnode)
            last = brnode.orelse
        last.extend(self.body(node.orfail))
        last.append(INCGRD)
        whilenode = While(pyTrue(), body, [])
        main = [Assign([pyName(node.unique_label)], Num(0))]
        if node.timeout is not None:
            main.append(Expr(pyCall(pyAttr("self", "_timer_start"))))
        main.append(whilenode)
        return concat_bodies(conds, main)

    def visit_ReturnStmt(self, node):
        if node.value is not None:
            value = self.visit(node.value)
        else:
            value = None
        ast = Return(value)
        return concat_bodies([value], [ast])

    def visit_DeleteStmt(self, node):
        targets = [self.visit(tgt) for tgt in node.targets]
        ast = Delete(targets)
        return concat_bodies(targets, [ast])

    def visit_YieldStmt(self, node):
        if node.value is not None:
            value = self.visit(node.value)
            ast = Expr(Yield(value))
            return concat_bodies([value], [ast])
        else:
            return [Expr(Yield(None))]

    def visit_YieldFromStmt(self, node):
        if node.value is not None:
            value = self.visit(node.value)
            ast = Expr(YieldFrom(value))
            return concat_bodies([value], [ast])
        else:
            return [Expr(YieldFrom(None))]

    def visit_WithStmt(self, node):
        items = []
        for item in node.items:
            context_expr = self.visit(item[0])
            if item[1] is not None:
                optional_vars = self.visit(item[1])
            else:
                optional_vars = None
            items.append(withitem(context_expr, optional_vars))
        body = self.body(node.body)
        ast = With(items, body)
        return concat_bodies([e.context_expr for e in items], [ast])

    def visit_RaiseStmt(self, node):
        ast = Raise(self.visit(node.expr), self.visit(node.cause))
        return concat_bodies((node.expr, node.cause), [ast])

    def visit_SimpleStmt(self, node):
        value = self.visit(node.expr)
        ast = Expr(value)
        return concat_bodies([value], [ast])

    def visit_BreakStmt(self, node):
        return [Break()]

    def visit_PassStmt(self, node):
        return [Pass()]

    def visit_ContinueStmt(self, node):
        return [Continue()]

    def visit_ImportStmt(self, node):
        names = []
        for item in node.items:
            if item.asname is None:
                names.append(alias(item.name.name, None))
            else:
                names.append(alias(item.name, item.asname.name))
        return [Import(names)]

    def visit_ImportFromStmt(self, node):
        names = []
        for item in node.items:
            if item.asname is None:
                names.append(alias(item.name.name, None))
            else:
                names.append(alias(item.name, item.asname.name))
        return [ImportFrom(node.module, names, node.level)]

    def visit_AssertStmt(self, node):
        expr = self.visit(node.expr)
        msg = self.visit(node.msg) if node.msg is not None else None
        ast = Assert(expr, msg)
        return concat_bodies([expr, msg], [ast])

    def visit_GlobalStmt(self, node):
        return [Global(node.names)]

    def visit_NonlocalStmt(self, node):
        return [Nonlocal(node.names)]

    def visit_OutputStmt(self, node):
        args = [self.visit(msg) for msg in node.message]
        keywords = []
        if node.level is not None:
            keywords.append(("level", self.visit(node.level)))
        if node.separator is not None:
            keywords.append(("sep", self.visit(node.separator)))
        ast = Expr(pyCall(func=pyAttr("self", "output"),
                          args=args, keywords=keywords))
        return concat_bodies(args, [ast])

    def visit_ResetStmt(self, node):
        blueprint = """
for attr in dir(self):
    if attr.find("{0}Event_") != -1:
        getattr(self, attr).clear()
"""
        if node.expr is None:
            typestr = ""
        else:
            typestr = str(node.expr.value)
        src = blueprint.format(typestr)
        return parse(src).body

    def history_stub(self, node):
        if node.record_history:
            return pyTrue()
        else:
            return pyNone()

    def visit_Event(self, node):
        return pyAttr(pyAttr("da", "pat"), node.type.__name__)

    def visit_EventHandler(self, node):
        stmts = self.visit_Function(node)
        stmts.append(Assign([pyAttr(node.name, "_labels")],
                            (pyNone() if node.labels is None else
                             pyCall(pyName("frozenset"),
                                    [Set([Str(l) for l in node.labels])]))))
        stmts.append(Assign([pyAttr(node.name, "_notlabels")],
                            (pyNone() if node.notlabels is None else
                             pyCall(pyName("frozenset"),
                                    [Set([Str(l) for l in node.notlabels])]))))
        return stmts

class PatternComprehensionGenerator(PythonGenerator):
    def __init__(self):
        super().__init__()
        # Set of freevars seen so far. Freevars after the first occurrence
        # needs to be unified:
        self.freevars = set()
        self.state_stack = []

    def push_state(self):
        self.state_stack.append(frozenset(self.freevars))

    def pop_state(self):
        s = self.state_stack.pop()
        self.freevars = set(s)

    def reset_state(self):
        self.freevars = set()

    def visit_FreePattern(self, node):
        conds = []
        if node.value is None:
            target = pyName("_")
        elif node.value in self.freevars:
            target = pyName(node.unique_name)
            conds = [pyCompare(target, Eq, self.visit(node.value))]
        else:
            target = self.visit(node.value)
            self.freevars.add(node.value)
        return target, conds

    def visit_BoundPattern(self, node):
        boundname = pyName(node.unique_name)
        targetname = self.visit(node.value)
        conast = pyCompare(boundname, Eq, targetname)
        return boundname, [conast]

    def visit_ConstantPattern(self, node):
        target = pyName(node.unique_name)
        compval = self.visit(node.value)
        return target, [pyCompare(target, Eq, compval)]

    def visit_TuplePattern(self, node):
        condition_list = []
        targets = []
        for elt in node.value:
            tgt, conds = self.visit(elt)
            targets.append(tgt)
            condition_list.extend(conds)
        if is_all_wildcards(targets):
            # Optimization: combine into one '_'
            return pyName('_'), []
        target = pyTuple(targets)
        return target, condition_list

    def visit_ListPattern(self, node):
        raise NotImplementedError(
            "Can not compile list pattern to comprehension.")

    def visit_PatternExpr(self, node):
        return self.visit(node.pattern)

    visit_LiteralPatternExpr = visit_PatternExpr



# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import logging
import argparse

from .api import entrypoint, DEFAULT_MASTER_PORT
from . import common
from .common import initialize_runtime_options

__version__ = common.__version__

if hasattr(sys, '_real_argv'):
    sys.argv[0] = sys._real_argv

def parseConfig(item):
    try:
        key, value = item.split('=')
        return key, value
    except ValueError:
        die("Invalid configuration format: %s" % item)

def parseArgs():
    LogLevelNames = [n.lower() for n in logging._nameToLevel]

    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')

    parser.add_argument('--iterations', type=int, default=1,
                        help="number of times to run the program, defaults to 1.")
    parser.add_argument("--no-log",
                        action="store_true", default=False,
                        help="if set, don't customize the root logger. "
                        "Useful if DistAlgo is run as a library .")
    parser.add_argument("-f", "--logfile", action="store_true", default=False,
                        help="creates a log file for this run.")
    parser.add_argument("--logfilename",
                        help="file name of the log file, defaults to appending"
                        "'.log' to the source file name.")
    parser.add_argument("--logdir")
    parser.add_argument("-L", "--logconsolelevel",
                        choices=LogLevelNames, default="info",
                        help="severity level of logging messages to print to "
                        "the console, defaults to 'info'.")
    parser.add_argument("-F", "--logfilelevel",
                        choices=LogLevelNames, default="debug",
                        help="severity level of logging messages to log to "
                        "the log file, defaults to 'debug'.")
    parser.add_argument("--pid-format",
                        choices=['short', 'long', 'full'], default='short',
                        help="sets the format of string representation of "
                        "process ids. 'short' prints the process class name "
                        "followed by the uid truncated to the last 5 hexdigits. "
                        "This is the default. 'long' prints the process class "
                        "name followed by the full 24 hexdigit untruncated uid. "
                        "'full' prints the full string representation of the "
                        "process id object. For named processes, both the "
                        "'short' and 'long' forms print the process name "
                        "in place of the uid.")
    parser.add_argument("-i", "--load-inc-module",
                        action="store_true", default=False,
                        help="if set, try to load the incrementalized "
                        "interface module.")
    parser.add_argument("-C", "--control-module-name", default=None,
                        help="name of the control module. If set, "
                        "results from the inc-module will be compared "
                        "against results from this module. Any mismatch will "
                        "raise IntrumentationError. Defaults to no control "
                        "module.")
    parser.add_argument("--inc-module-name",
                        help="name of the incrementalized interface module, "
                        "defaults to source module name + '_inc'. ")
    parser.add_argument("-H", "--hostname", default='localhost',
                        help="hostname for binding network sockets, "
                        "defaults to 'localhost'. This option is ignored "
                        "if '--nodename' is not set.")
    parser.add_argument("-p", "--port", type=int, default=None,
                        help="port number for binding network sockets. "
                        "Default is bind to any random available port. "
                        "This option is ignored if '--nodename' is not set.")
    parser.add_argument("-n", "--nodename", default="",
                        help="mnemonic name for this DistAlgo node process. "
                        "Default is no name. ")
    parser.add_argument("-R", "--peer", default="",
                        help="address of an existing node, used for bootstrapping "
                        "this node. An address should be of the "
                        "form HOSTNAME[:PORT], where 'PORT' defaults to the "
                        "value of '--default-master-port' if omitted. "
                        "This option is ignored if '--nodename' is not set, "
                        "or if '--master' is set.")
    parser.add_argument("--default-master-port", type=int, default=15000,
                        help="default port number for connecting to remote "
                        "nodes if the port number is not explicitly given. "
                        "Default value is 15000. "
                        "This option is ignored if '--nodename' is not set.")
    parser.add_argument("--cookie", default=None,
                        help="a string for authentication of peers. "
                        "All peer processes participating in message passing "
                        "must have matching cookies. "
                        "If this option is not set, but the file "
                        "'${HOME}/.da.cookie' exists, then the contents of "
                        "that file will be used as the "
                        "cookie. If this option is not set and the file "
                        "'${HOME}/.da.cookie' does not exist, then peer "
                        "authentication will be disabled for this node.")
    parser.add_argument('--message-buffer-size', type=int, default=(4 * 1024),
                        help="size in bytes of the buffers used to serialize "
                        "messages. The serialized(pickled) size of any DistAlgo "
                        "message must be smaller than this value in order for the "
                        "system to be able to send it "
                        "across address space boundaries. Default value is 4KB.")
    parser.add_argument("--tcp-dont-reuse-addr",
                        help="if set, the system will not bind to TCP ports "
                        "that are in the 'TIME_WAIT' state. ",
                        action="store_true", default=False)
    parser.add_argument("-M", "--master",
                        help="if set, the system will not try to bootstrap "
                        "on startup. ",
                        action="store_true", default=False)
    parser.add_argument("-D", "--idle",
                        help="if set, this node will run as an idle node, "
                        "i.e. it will not execute the `main` method defined "
                        "in the main module. ",
                        action="store_true", default=False)
    parser.add_argument("-r", "--recompile", dest="recompile",
                        help="force recompile DistAlgo source file. ",
                        action="store_true", default=False)
    parser.add_argument("-c", "--compiler-flags", default="",
                        help="flags to pass to the compiler, if (re)compiling "
                        "is required.")
    parser.add_argument("-o", "--config", default=[], nargs='*',
                        help="sets runtime configuration variables, overrides "
                        "configurations declared in the program source.")
    parser.add_argument("--start-method", default=None, choices=['fork', 'spawn'],
                        help="choose the semantics for creating child process."
                        " 'fork' is the default method on UNIX-like systems,"
                        " 'spawn' is the default method on Windows systems.")
    parser.add_argument("-I", "--default-proc-impl", default='process',
                        choices=['process', 'thread'],
                        help="choose the default implementation for running "
                        " DistAlgo processes."
                        " 'process' uses OS processes,"
                        " 'thread' uses OS threads.")
    parser.add_argument("-v", "--version", action="version", version=__version__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-m", "--module", default=None, nargs='+',
                        help="name of a DistAlgo module that will be run as "
                       "the main module. If this argument is specified, "
                       "all command line options after this point will be "
                       "passed to the specified module in `sys.argv`.")
    group.add_argument("-B", "--help-builtins", action='store_true', default=False,
                       help="print a list of DistAlgo built-in functions and "
                       "exit.")
    group.add_argument("file", nargs='?',
                        help="DistAlgo source file to run.")
    parser.add_argument("args", nargs=argparse.REMAINDER,
                        help="arguments passed to program in sys.argv[1:].")

    args = parser.parse_args()
    if args.help_builtins:
        help_builtins()
        return 0
    elif not args.idle and args.module is None and args.file is None:
        parser.print_usage()
        return 1
    args.config = dict(parseConfig(item) for item in args.config)
    if args.cookie is not None:
        args.cookie = args.cookie.encode()
    return args

def help_builtins():
    from inspect import signature
    print("======= DistAlgo Builtin Functions =======\n")
    for fname in sorted(common.builtin_registry):
        func = common.builtin_registry[fname]
        sig = ", ".join([arg for arg in signature(func).parameters
                         if arg != 'self'])
        print("{}({}):\n\t{}\n".format(fname, sig, func.__doc__))

def libmain():
    """Main program entry point.

    Parses command line options, sets up global variables, and calls the 'main'
    function of the DistAlgo program.

    """
    args = parseArgs()
    if isinstance(args, int):
        return args
    else:
        initialize_runtime_options(args.__dict__)
        return entrypoint()

def die(mesg = None):
    if mesg != None:
        sys.stderr.write(mesg + "\n")
    sys.exit(1)

if __name__ == '__main__':
    sys.exit(libmain())



# -*- generated by 1.0.12 -*-
import da
PatternExpr_188 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.FreePattern(None)])
PatternExpr_194 = da.pat.BoundPattern('_BoundPattern195_')
PatternExpr_218 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('aborting')])
PatternExpr_225 = da.pat.BoundPattern('_BoundPattern226_')
PatternExpr_246 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready')])
PatternExpr_253 = da.pat.FreePattern('c')
PatternExpr_286 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready')])
PatternExpr_293 = da.pat.BoundPattern('_BoundPattern294_')
PatternExpr_319 = da.pat.ConstantPattern('done')
PatternExpr_323 = da.pat.BoundPattern('_BoundPattern324_')
PatternExpr_227 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern233_')]), da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('aborting')])])
PatternExpr_295 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern301_')]), da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready')])])
PatternExpr_325 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern331_')]), da.pat.ConstantPattern('done')])
PatternExpr_373 = da.pat.ConstantPattern('prepare')
PatternExpr_377 = da.pat.FreePattern('coord')
PatternExpr_403 = da.pat.ConstantPattern('abort')
PatternExpr_412 = da.pat.ConstantPattern('commit')
PatternExpr_416 = da.pat.FreePattern('coord')
_config_object = {}
import sys
from random import randint

class Coordinator(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._CoordinatorReceivedEvent_0 = []
        self._CoordinatorReceivedEvent_1 = []
        self._CoordinatorReceivedEvent_2 = []
        self._CoordinatorReceivedEvent_3 = []
        self._CoordinatorReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_0', PatternExpr_188, sources=[PatternExpr_194], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_1', PatternExpr_218, sources=[PatternExpr_225], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_2', PatternExpr_246, sources=[PatternExpr_253], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_3', PatternExpr_286, sources=[PatternExpr_293], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_4', PatternExpr_319, sources=[PatternExpr_323], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, cohorts, **rest_515):
        super().setup(cohorts=cohorts, **rest_515)
        self._state.cohorts = cohorts
        pass

    def run(self):
        self.to_commit()
        self.output('terminating')

    def to_commit(self):
        self.send('prepare', to=self._state.cohorts)
        super()._label('_st_label_179', block=False)
        c = None

        def UniversalOpExpr_180():
            nonlocal c
            for c in self._state.cohorts:

                def ExistentialOpExpr_186(c):
                    for (_, (_, _, _BoundPattern202_), (_ConstantPattern204_, _)) in self._CoordinatorReceivedEvent_0:
                        if (_BoundPattern202_ == c):
                            if (_ConstantPattern204_ == 'vote'):
                                if True:
                                    return True
                    return False
                if (not ExistentialOpExpr_186(c=c)):
                    return False
            return True
        _st_label_179 = 0
        while (_st_label_179 == 0):
            _st_label_179 += 1
            if UniversalOpExpr_180():
                _st_label_179 += 1
            else:
                super()._label('_st_label_179', block=True)
                _st_label_179 -= 1
        c = None

        def ExistentialOpExpr_211():
            nonlocal c
            for c in self._state.cohorts:
                if PatternExpr_227.match_iter(self._CoordinatorReceivedEvent_1, _BoundPattern233_=c, SELF_ID=self._id):
                    return True
            return False
        if ExistentialOpExpr_211():
            s = {c for c in self._state.cohorts for (_, (_, _, _FreePattern261_), (_ConstantPattern263_, _ConstantPattern265_)) in self._CoordinatorReceivedEvent_2 if (_FreePattern261_ == c) if (_ConstantPattern263_ == 'vote') if (_ConstantPattern265_ == 'ready')}
            self.send('abort', to=s)
            self.abort()
        c = None

        def UniversalOpExpr_279():
            nonlocal c
            for c in self._state.cohorts:
                if (not PatternExpr_295.match_iter(self._CoordinatorReceivedEvent_3, _BoundPattern301_=c, SELF_ID=self._id)):
                    return False
            return True
        if UniversalOpExpr_279():
            self.send('commit', to=self._state.cohorts)
            super()._label('_st_label_311', block=False)
            c = None

            def UniversalOpExpr_312():
                nonlocal c
                for c in self._state.cohorts:
                    if (not PatternExpr_325.match_iter(self._CoordinatorReceivedEvent_4, _BoundPattern331_=c, SELF_ID=self._id)):
                        return False
                return True
            _st_label_311 = 0
            while (_st_label_311 == 0):
                _st_label_311 += 1
                if UniversalOpExpr_312():
                    _st_label_311 += 1
                else:
                    super()._label('_st_label_311', block=True)
                    _st_label_311 -= 1
            self.commit()

    def abort(self):
        self.output('abort')

    def commit(self):
        self.output('commit')

class Cohort(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_0', PatternExpr_373, sources=[PatternExpr_377], destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_372]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_1', PatternExpr_403, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_402]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_2', PatternExpr_412, sources=[PatternExpr_416], destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_411])])

    def setup(self, failure_rate, **rest_515):
        super().setup(failure_rate=failure_rate, **rest_515)
        self._state.failure_rate = failure_rate
        self._state.terminate = False

    def run(self):
        super()._label('_st_label_455', block=False)
        _st_label_455 = 0
        while (_st_label_455 == 0):
            _st_label_455 += 1
            if self._state.terminate:
                _st_label_455 += 1
            else:
                super()._label('_st_label_455', block=True)
                _st_label_455 -= 1

    def prepared(self):
        return (randint(0, 100) > self._state.failure_rate)

    def ready(self):
        self.output('ready')

    def abort(self):
        self.output('abort')
        self._state.terminate = True

    def commit(self):
        self.output('commit')
        self._state.terminate = True

    def _Cohort_handler_372(self, coord):
        if self.prepared():
            self.send(('vote', 'ready'), to=coord)
            self.ready()
        else:
            self.send(('vote', 'aborting'), to=coord)
            self.abort()
    _Cohort_handler_372._labels = None
    _Cohort_handler_372._notlabels = None

    def _Cohort_handler_402(self):
        self.abort()
    _Cohort_handler_402._labels = None
    _Cohort_handler_402._notlabels = None

    def _Cohort_handler_411(self, coord):
        self.send('done', to=coord)
        self.commit()
    _Cohort_handler_411._labels = None
    _Cohort_handler_411._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        ncohorts = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        fail_rate = (int(sys.argv[2]) if (len(sys.argv) > 2) else 10)
        cohorts = self.new(Cohort, (fail_rate,), num=ncohorts)
        coordinators = self.new(Coordinator, (cohorts,), num=1)
        self._start((cohorts | coordinators))



# -*- generated by 1.0.12 -*-
import da
PatternExpr_235 = da.pat.TuplePattern([da.pat.ConstantPattern('ack'), da.pat.FreePattern('c2'), da.pat.BoundPattern('_BoundPattern239_')])
PatternExpr_279 = da.pat.TuplePattern([da.pat.ConstantPattern('request'), da.pat.FreePattern('c2'), da.pat.FreePattern('p')])
PatternExpr_305 = da.pat.TuplePattern([da.pat.ConstantPattern('release'), da.pat.FreePattern(None), da.pat.FreePattern('p')])
PatternExpr_364 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_369 = da.pat.BoundPattern('_BoundPattern371_')
PatternExpr_372 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern378_')]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
PatternExpr_454 = da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern457_')])
PatternExpr_460 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern470_')])])
_config_object = {'channel': 'fifo', 'clock': 'lamport'}
import sys

class P(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._PReceivedEvent_0 = []
        self._PReceivedEvent_3 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_0', PatternExpr_235, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_1', PatternExpr_279, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_278]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_2', PatternExpr_305, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_304]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_3', PatternExpr_364, sources=[PatternExpr_369], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, s, nrequests, **rest_476):
        super().setup(s=s, nrequests=nrequests, **rest_476)
        self._state.s = s
        self._state.nrequests = nrequests
        self._state.q = set()

    def run(self):

        def task():
            self.output('in cs')
        for i in range(self._state.nrequests):
            self.mutex(task)
        self.send(('done', self._id), to=self.parent())
        super()._label('_st_label_361', block=False)
        _st_label_361 = 0
        while (_st_label_361 == 0):
            _st_label_361 += 1
            if PatternExpr_372.match_iter(self._PReceivedEvent_3, _BoundPattern378_=self.parent(), SELF_ID=self._id):
                _st_label_361 += 1
            else:
                super()._label('_st_label_361', block=True)
                _st_label_361 -= 1
        self.output('terminating')

    def mutex(self, task):
        super()._label('request', block=False)
        c = self.logical_clock()
        self.send(('request', c, self._id), to=self._state.s)
        self._state.q.add(('request', c, self._id))
        super()._label('_st_label_200', block=False)
        p = c2 = None

        def UniversalOpExpr_202():
            nonlocal p, c2
            for (_ConstantPattern206_, c2, p) in self._state.q:
                if (_ConstantPattern206_ == 'request'):
                    if (not (((c2, p) == (c, self._id)) or ((c, self._id) < (c2, p)))):
                        return False
            return True
        p = c2 = None

        def UniversalOpExpr_228():
            nonlocal p, c2
            for p in self._state.s:

                def ExistentialOpExpr_233(p):
                    nonlocal c2
                    for (_, _, (_ConstantPattern250_, c2, _BoundPattern253_)) in self._PReceivedEvent_0:
                        if (_ConstantPattern250_ == 'ack'):
                            if (_BoundPattern253_ == p):
                                if (c2 > c):
                                    return True
                    return False
                if (not ExistentialOpExpr_233(p=p)):
                    return False
            return True
        _st_label_200 = 0
        while (_st_label_200 == 0):
            _st_label_200 += 1
            if (UniversalOpExpr_202() and UniversalOpExpr_228()):
                _st_label_200 += 1
            else:
                super()._label('_st_label_200', block=True)
                _st_label_200 -= 1
        super()._label('critical_section', block=False)
        task()
        super()._label('release', block=False)
        self._state.q.remove(('request', c, self._id))
        self.send(('release', self.logical_clock(), self._id), to=self._state.s)

    def _P_handler_278(self, c2, p):
        self._state.q.add(('request', c2, p))
        self.send(('ack', self.logical_clock(), self._id), to=p)
    _P_handler_278._labels = None
    _P_handler_278._notlabels = None

    def _P_handler_304(self, p):
        for x in {('request', c, p) for (_ConstantPattern321_, c, _BoundPattern324_) in self._state.q if (_ConstantPattern321_ == 'request') if (_BoundPattern324_ == p)}:
            self._state.q.remove(x)
            break
    _P_handler_304._labels = None
    _P_handler_304._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._Node_ReceivedEvent_0 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_Node_ReceivedEvent_0', PatternExpr_454, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def run(self):
        nprocs = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        nrequests = (int(sys.argv[2]) if (len(sys.argv) > 2) else 1)
        ps = self.new(P, num=nprocs)
        for p in ps:
            self._setup(p, ((ps - {p}), nrequests))
        self._start(ps)
        super()._label('_st_label_446', block=False)
        p = None

        def UniversalOpExpr_447():
            nonlocal p
            for p in ps:
                if (not PatternExpr_460.match_iter(self._Node_ReceivedEvent_0, _BoundPattern470_=p)):
                    return False
            return True
        _st_label_446 = 0
        while (_st_label_446 == 0):
            _st_label_446 += 1
            if UniversalOpExpr_447():
                _st_label_446 += 1
            else:
                super()._label('_st_label_446', block=True)
                _st_label_446 -= 1
        self.send(('done',), to=ps)



# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import io
import sys

import da
from da.tools.unparse import Unparser

DB_ERROR = 0
DB_WARN = 1
DB_INFO = 2
DB_DEBUG =3
Debug = DB_INFO

VERSION_HEADER = "# -*- generated by {} -*-"

def to_source(tree):
    textbuf = io.StringIO(newline='')
    textbuf.write(VERSION_HEADER.format(da.__version__))
    Unparser(tree, textbuf)
    return textbuf.getvalue()

def to_file(tree, fd):
    fd.write(VERSION_HEADER.format(da.__version__))
    return Unparser(tree, fd).counter

def set_debug_level(level):
    global Debug
    if is_valid_debug_level(level):
        Debug = level

def is_valid_debug_level(level):
    return type(level) is int and DB_ERROR <= level and level <= DB_DEBUG

# Common utility functions

def printe(mesg, lineno=0, col_offset=0, filename="", outfd=sys.stderr):
    if Debug >= DB_ERROR:
        fs = "%s:%d:%d: error: %s"
        print(fs % (filename, lineno, col_offset, mesg), file=outfd)

def printw(mesg, lineno=0, col_offset=0, filename="", outfd=sys.stderr):
    if Debug >= DB_WARN:
        fs = "%s:%d:%d: warning: %s"
        print(fs % (filename, lineno, col_offset, mesg), file=outfd)

def printd(mesg, lineno=0, col_offset=0, filename="", outfd=sys.stderr):
    if Debug >= DB_DEBUG:
        fs = "%s:%d:%d: DEBUG: %s"
        print(fs % (filename, lineno, col_offset, mesg), file=outfd)

def printi(mesg, lineno=0, col_offset=0, filename="", outfd=sys.stdout):
    if Debug >= DB_INFO:
        fs = "%s:%d:%d: %s"
        print(fs % (filename, lineno, col_offset, mesg), file=outfd)

class Namespace:
    """A simple container for storing arbitrary attributes."""
    pass

class OptionsManager:
    def __init__(self, cmdline_args, module_args, default=False):
        self.cmdline_args = cmdline_args
        self.module_args = module_args
        self.default = default

    def __getattribute__(self, option):
        if option in {'cmdline_args', 'module_args', 'default'}:
            return super().__getattribute__(option)

        if hasattr(self.cmdline_args, option):
            return getattr(self.cmdline_args, option)
        elif hasattr(self.module_args, option):
            return getattr(self.module_args, option)
        else:
            return self.default



import os
import distutils
import setuptools
import subprocess
import setuptools.command.build_py
import setuptools.command.sdist
from distutils.core import setup
from distutils.cmd import Command

import da

class CompileDACommand(Command):
    """A custom command to run `dac` on all .da files under da/."""

    description = 'build all DistAlgo modules under da/.'
    user_options = [
        # The format is (long option, short option, description).
        ('compiler-flags', 'F', 'options for the compiler'),
    ]

    def initialize_options(self):
        """Set default values for options."""
        # Each user option must be listed here with their default value.
        self.compiler_flags = []

    def finalize_options(self):
        """Post-process options."""
        pass

    def run(self):
        """Run command."""
        from da import compiler
        exroot = os.path.join(os.getcwd(), 'da/')
        for root, _, files in os.walk(exroot):
            for filename in files:
                if filename.endswith('.da'):
                    target = os.path.join(root, filename)
                    self.announce('Compiling {}...'.format(target),
                                  level=distutils.log.INFO)
                    compiler.ui.main(self.compiler_flags + [str(target)])

class CompileDocCommand(Command):
    """A custom command to run `pdflatex` on 'doc/language.tex'."""

    description = "generate 'doc/language.pdf' from 'doc/language.tex'."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        rootdir = os.getcwd()
        os.chdir('./doc')
        command = ['pdflatex', 'language.tex']
        self.announce('Running command: {}'.format(command))
        subprocess.check_call(command)
        os.chdir(rootdir)

class DABuildCommand(setuptools.command.build_py.build_py):
    """Auto build all examples before packaging."""

    def run(self):
        self.run_command('compile_modules')
        super().run()

class DASdistCommand(setuptools.command.sdist.sdist):
    """Generate doc/language.pdf before packaging."""

    def run(self):
        self.run_command('compile_modules')
        self.run_command('genpdf')
        super().run()

setup(name = "pyDistAlgo",
      version = da.__version__,
      url = "https://github.com/DistAlgo/distalgo",
      description = "A high-level language for distributed algorithms.",
      author = "bolin",
      author_email = "bolin@cs.stonybrook.edu",
      classifiers = [
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Compilers',
      ],
      packages = setuptools.find_packages(),
      include_package_data = True,
      cmdclass = {
          'genpdf'          : CompileDocCommand,
          'compile_modules' : CompileDACommand,
          'build_py'        : DABuildCommand,
          'sdist'           : DASdistCommand,
      }
)



# -*- generated by 1.0.12 -*-
import da
_config_object = {}

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        return 0
